# coding: utf-8
import os
import numpy as np

from ..utils import handleKeyError
from ..utils import flatten_dual
from ..utils import ItemsetTreeDOTexporter

ITEM_MINING_METHODS = ["all", "closed"]

def create_one_hot(data):
    """
    Create the one-hot binary matrix.
    @params data     : (list) Each element of data (data[i]) have variable length items.
    @return one_hot  : shape=(num_data,num_unique)
        one_hot[n][i]=1 means data[n] contains idx2data[i]
    @return idx2data : Dictionary from index to original data.
    """
    unique_data = sorted(list(set(flatten_dual(data))))
    num_unique = len(unique_data)
    num_data = len(data)
    data2idx = dict(zip(unique_data, range(num_unique)))
    one_hot = np.zeros(shape=(num_data,num_unique), dtype=np.int32)
    for i,row in enumerate(data):
        one_hot[i, np.asarray([data2idx[e] for e in row], dtype=int)]=1
    idx2data = dict(zip(range(num_unique), unique_data))
    return one_hot, idx2data

class Node():
    """ Node for Tree structure.
    @params database  :
    @params itemset   :
    @params num_items :
    @params freq      :
    @params tail      :
    @params threshold :
    """
    def __init__(self, itemset, freq, tail):
        self.itemset = itemset
        self.freq = freq # (=len(database))
        self.tail = tail
        self.sons = []

    def _recurse_all(self, database, threshold, num_items):
        """ Find ALL closed Itemsets. """
        for i in range(self.tail+1, num_items):
            next_itemset = self.itemset + [i]
            next_data = database[database[:,i]==1,:]
            freq = len(next_data)
            if freq >= threshold:
                son = Node(itemset=next_itemset, freq=freq, tail=i)
                son._recurse_all(next_data, threshold, num_items)
                self.sons.append(son)

    def _recurse_closed(self, database, threshold, num_items):
        """ Find ONLY closed Itemsets. """
        for i in range(self.tail+1, num_items):
            next_data = database[database[:,i]==1,:]
            freq = len(next_data)
            if freq >= threshold:
                add_itemset = i+np.where(np.all(next_data[:,i:], axis=0))[0]
                next_itemset = self.itemset + add_itemset.tolist()
                son = Node(itemset=next_itemset, freq=freq, tail=max(add_itemset))
                son._recurse_closed(next_data, threshold, num_items)
                self.sons.append(son)

class FrequentSet():
    def __init__(self, threshold):
        self.tree = None
        self.threshold = threshold
        self.freq_sets = []

    def fit(self, database, method="closed"):
        """
        @param database: Binary Matrix. shape=(num_transactions, num_items)
        """
        method = method.lower()
        handleKeyError(lst=ITEM_MINING_METHODS, method=method)
        num_transactions, num_items = database.shape
        self.tree = Node(itemset=[], freq=num_transactions, tail=-1)
        self.tree.__getattribute__({
            "all"    : "_recurse_all",
            "closed" : "_recurse_closed",
        }[method]).__call__(database, self.threshold, num_items)
        self.num_items = num_items
        self.all = self.get_itemsets(self.tree)

    def get_itemsets(self, node):
        freq_sets = [node.itemset]
        for son in node.sons:
            freq_sets.extend(self.get_itemsets(node=son))
        return freq_sets

    def export_graphviz(self, out_file=None, feature_names=None,
                        class_names=None, cmap="jet", filled=True,
                        rounded=True, precision=3):
        if class_names is None:
            class_names = np.arange(self.num_items)
        exporter = ItemsetTreeDOTexporter(
            cmap=cmap, class_names=class_names,
            filled=filled, rounded=rounded, precision=precision
        )
        if out_file is not None:
            ext = os.path.splitext(os.path.basename(out_file))[-1]
            if ext==".png":
                return exporter.write_png(self.tree, path=out_file)
            elif ext==".dot":
                return exporter.export(self.tree, out_file=out_file)
        return exporter.export(self.tree, out_file=None)



# coding: utf-8
import re
import copy
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as colors
from matplotlib._cm import datad
from mpl_toolkits.mplot3d import Axes3D

from .generic_utils import handleKeyError
from .metric_utils import silhouette_samples
from .metric_utils import silhouette_score

ALL_CMAP_NAMES = [name for name in datad.keys() if not re.search(r".*_r", name, re.IGNORECASE)]
COLOR_TYPES = ["rgba", "rgb", "hex"]

def rgb2hex(rgb, max_val=1):
    return "#"+"".join([format(int(255/max_val*e), '02x') for e in rgb]).upper()

def rgba2rgb(rgba, max_val=1):
    alpha = rgba[-1]
    rgb = rgba[:-1]
    type_ = int if max_val==255 else float
    # compute the color as alpha against white
    return tuple([type_(alpha*e+(1-alpha)*max_val) for e in rgb])

def hex2rgb(hex, max_val=1):
    tuple([int(hex[-6:][i*2:(i+1)*2], 16)/255*max_val for i in range(3)])

def chooseTextColor(rgb, ctype="rgb", max_val=1):
    # Ref: WCAG (https://www.w3.org/TR/WCAG20/)
    R,G,B = [e/max_val for e in rgb]
    # Relative Brightness BackGround.
    Lbg = 0.2126*R + 0.7152*G + 0.0722*B

    Lw = 1 # Relative Brightness of White
    Lb = 0 # Relative Brightness of Black

    Cw = (Lw + 0.05) / (Lbg + 0.05)
    Cb = (Lbg + 0.05) / (Lb + 0.05)
    return (0,0,0) if Cb>Cw else (max_val,max_val,max_val)

def mk_color_dict(keys, cmap="jet", reverse=False, max_val=1, ctype="rgba"):
    """
    @params keys    :
    @params cmap    :
    @params reverse :
    @params max_val :
    @params ctype   : rgb, rgba, hex
    """
    handleKeyError(lst=COLOR_TYPES, ctype=ctype)
    if not isinstance(cmap, colors.Colormap):
        # If `cmap` is string.
        handleKeyError(lst=ALL_CMAP_NAMES, cmap=cmap)
        if reverse:
            cmap += "_r"
        cmap = plt.get_cmap(cmap)

    if isinstance(keys, int):
        N  = keys-1
        color_dict_rgba = {n: cmap(n/N) for n in range(keys)}
    else:
        unique_keys = sorted(list(set(keys)))
        N = len(unique_keys)-1
        color_dict_rgba = {
            key: tuple([
                max_val*e if i<3 else e for i,e in enumerate(cmap(n/N))
            ]) for n,key in enumerate(unique_keys)
        }
    if ctype=="rgba":
        return color_dict_rgba

    color_dict_rgb = {k:rgba2rgb(v, max_val=max_val) for k,v in color_dict_rgba.items()}
    if ctype=="rgb":
        return color_dict_rgb

    if ctype=="hex":
        color_dict_hex = {k:rgb2hex(v[:-1],max_val=max_val) for k,v in color_dict_rgb.items()}
    return color_dict_hex

def measureCanvas(nfigs, ncols_max=2, figinches=(6,4)):
    """
    @params nfigs    : (int) Total number of figures.
    @params ncols_max: (int) The number of columns.
    @params figinches: (tuple) The figure size for 1 plot.
    """
    if nfigs>=ncols_max:
        ncols=ncols_max
        nrows=(nfigs-1)//ncols_max+1
    else:
        ncols=nfigs
        nrows=1
    w,h = figinches
    figsize=(w*ncols,h*nrows)
    return (ncols, nrows, figsize)

def galleryplot(func, argnames, iterator, ncols=5, figsize=(4,4), sharex="none", sharey="none", **kwargs):

    """Plot many figures as a gallery.

    Parameters
    ----------
    @param func         : function which plots each image. NOTE: they must recieve the Axes object.
    @param argnames     : keys of one or more arguments. These are given to `func`.
    @param iterator     : iterator which has one or more values, or zip object of them.
    @param ncols        : number of images arranged horizontally (x-axis).
    @param figsize      : (Widht, Height). image size of each image.
    @param sharex,sharey: bool or {'none', 'all', 'row', 'col'}
    @return fig         : class `matplotlib.figure.Figure` object.
    @return axes        : Axes object or array of Axes objects.

    Examples
    --------
    [1-argument]
    >>> def sample(ax, color):
            ax.scatter(0,0,color=color,s=500)
            return ax
    >>> colors=["red","blue","green","black"]
    >>> fig, axes = galleryplot(func=sample, argnames="color", iterator=colors, ncols=4)
    >>> plt.tight_layout()
    >>> plt.show()

    [multi-args]
    >>> def sample(ax, color, marker):
        ax.scatter(0,0,color=color,s=500, marker=marker)
        return ax
    >>> colors  = ["red","blue","green","black"]
    >>> markers = ["o","s","x","^"]
    >>> fig, axes = galleryplot(func=sample2, argnames=["color","marker"], iterator=zip(colors,markers), ncols=4)
    >>> plt.tight_layout()
    >>> plt.show()
    """
    iter_for_count = copy.deepcopy(iterator)
    nfigs = len([e for e in iter_for_count])
    ncols, nrows, figsize = measureCanvas(nfigs, ncols_max=ncols, figinches=figsize)

    fig, axes = plt.subplots(nrows=nrows, ncols=ncols, sharex=sharex, sharey=sharey, figsize=figsize)
    axes = axes if axes.ndim>1 else axes.reshape(1,-1)
    for i,vals in enumerate(iterator):
        kwargs = {argnames:vals} if isinstance(argnames, str) else dict(zip(argnames,vals))
        ax = func(ax=axes[i//ncols, i%ncols], **kwargs)

    return fig, axes

def plot_2D(*data, ax=None, figsize=(4,4), **plotargs):
    """ visualize 2-dimensional data. """
    if ax is None: fig, ax = plt.subplots(figsize=figsize)

    if len(data)==1:
        x = data[0]
        y = np.zeros(shape=len(x))
        label = False
    elif len(data)==2:
        x,y = data
        label = True
    else:
        raise ValueError(f"When passing data, it must contain 1 (only data) or 2 (data, class_label) items. However, it contains {len(data)} items.")
    n_cls = len(np.unique(y))
    for cls in np.unique(y):
        ax.scatter(x[y==cls, 0], x[y==cls, 1], label=f"cls.{cls}" if label else None, color=cm.jet(cls/n_cls))
    ax.set_xlabel("$x$", fontsize=14), ax.set_ylabel("$y$", fontsize=14)
    return ax

def objVSexp(obj, *exp, **options):
    """ Visualize 'Objective variable' vs. 'Explanatory variable' plot.
    @params
    @params var_type: str, 'continuous' or 'discrete' (default='continuous')
    """
    n_features = len(exp)
    if n_features==0 or n_features>3:
        raise ValueError(f"The number of explanatory variables should be 1 or 2, but {n_features}")

    var_type = options.pop("var_type", "continuous")
    figsize  = options.pop("figsize", None)
    ax       = options.pop("ax", None)
    color    = options.pop("color", "#722f37")
    alpha    = options.pop("alpha", .2)
    handleKeyError(["continuous", "discrete"], var_type=var_type)

    if ax is None:
        fig = plt.figure(figsize=figsize)
        projection = None if n_features==1 else "3d"
        ax = fig.add_subplot(1,1,1,projection=projection)

    if n_features==1:
        if var_type=="continuous":
            ax.scatter(*exp, obj, color=color, alpha=alpha)
        else:
            sns.boxplot(*exp, obj, ax=ax)
    else:
        ax.scatter(*exp, obj, color=color, alpha=alpha)

    return ax

def silhouette_plot(X, labels, centers, axes=None, set_style=True):
    """Silhouette Plotting.
    @params X       : Input data.               shape=(n_samples, n_features)
    @params labels  : Current label assignment. shape=(n_samples,)
    @params centers : The current centers.      shape=(n_clusters, n_features)
    @params axes    : (axL, axR) if `n_features` <= 3 else ax
    """
    n_features = X.shape[1]
    if n_features <= 3:
        if axes is None:
            fig = plt.figure(figsize=(18, 7))
            projection = "3d" if n_features==3 else None
            axes = (fig.add_subplot(1,2,1), fig.add_subplot(1,2,2,projection=projection))
        ax1 = _left_silhouette_plot(X, labels, ax=axes[0], set_style=set_style)
        ax2 = _right_silhouette_plot(X, labels, centers, ax=axes[1], set_style=set_style)
        axes = (ax1, ax2)
    else:
        if axes is None:
            fig, axes = plt.subplots()
            fig.set_size_inches(18, 7)
        axes = _left_silhouette_plot(X, labels, ax=axes, set_style=set_style)

    if set_style:
        plt.suptitle((f"Silhouette analysis"), fontsize=14, fontweight='bold')
    if axes is None:
        plt.show()
    else:
        return axes

def _left_silhouette_plot(X, labels, ax=None, set_style=True):
    if ax is None:
        fig, ax = plt.subplots()
        fig.set_size_inches(9, 7)

    silhouette_values = silhouette_samples(X, labels)
    silhouette_avg = np.mean(silhouette_values)

    y_lower = 10
    uniq_labels = np.unique(labels)
    n_clusters = len(uniq_labels)
    for i,k in enumerate(uniq_labels):
        Si_k = np.sort(silhouette_values[labels==k])
        Nk = Si_k.shape[0]
        y_upper = y_lower + Nk

        color = cm.hsv(float(i) / n_clusters)
        ax.fill_betweenx(np.arange(y_lower, y_upper), 0, Si_k,
                          facecolor=color, edgecolor=color, alpha=0.7)
        ax.text(-0.05, y_lower + 0.5 * Nk, str(i))
        y_lower = y_upper + 10

    ax.set_xlim([-0.1, 1])
    ax.set_ylim([0, len(X) + (n_clusters + 1) * 10])
    if set_style:
        ax.set_title("The silhouette plot for the various clusters.")
        ax.set_xlabel("The silhouette coefficient values")
        ax.set_ylabel("Cluster label")
        ax.axvline(x=silhouette_avg, color="black", linestyle="--", label="Average")
        ax.set_yticks([])  # Clear the yaxis labels / ticks
        ax.set_xticks([-0.1, 0, 0.2, 0.4, 0.6, 0.8, 1])
        ax.legend()
    return ax

def _right_silhouette_plot(X, labels, centers, ax=None, set_style=True):
    n_features = X.shape[1]
    if ax is None:
        fig = plt.figure(figsize=(9, 7))
        projection = "3d" if n_features==3 else None
        ax = fig.add_subplot(1,1,1,projection=projection)

    colors = cm.hsv(labels.astype(float) / (np.max(labels)+1.))
    if n_features == 1:
        X_zeros = np.zeros_like(X)
        C_zeros = np.zeros_like(centers)
        ax.scatter(X,       X_zeros, marker='.', s=30, lw=0, alpha=0.7, c=colors, edgecolor='k')
        ax.scatter(centers, C_zeros, marker='o', s=200, alpha=1, c="white", edgecolor='k')
        for i,c in enumerate(centers):
            ax.scatter(c, 0, marker='$%d$' % i, s=50, alpha=1, edgecolor='k')
        plt.tick_params(labelleft=False, left=False)
    elif n_features == 2:
        ax.scatter(X[:,0], X[:,1], marker='.', s=30, lw=0, alpha=0.7, c=colors, edgecolor='k')
        ax.scatter(centers[:,0], centers[:,1], marker='o', s=200, alpha=1, c="white", edgecolor='k')
        for i,c in enumerate(centers):
            ax.scatter(c[0], c[1], marker='$%d$' % i, s=50, alpha=1, edgecolor='k')
    elif n_features==3:
        ax.scatter(X[:,0], X[:,1], X[:,2], marker='.', s=60, lw=0, alpha=0.7, c=colors, edgecolor='k')
        ax.scatter(centers[:,0], centers[:,1], centers[:,2], marker='o', s=200, alpha=1, c="white", edgecolor='k')
        for i,c in enumerate(centers):
            ax.scatter(c[0], c[1], c[2], marker='$%d$' % i, s=50, alpha=1, edgecolor='k')
    else:
        raise ValueError(f"`X.shape[1]` should be smaller than or equal to 3, but got {n_features}")
    if set_style:
        ax.set_title("The visualization of the clustered data.")
        ax.set_xlabel("Feature space for the 1st feature")
        if n_features>=2: ax.set_ylabel("Feature space for the 2nd feature")
        if n_features==3: ax.set_zlabel("Feature space for the 3rd feature")
    return ax

def plot_CDF(data, ax=None, reverse=False, plot=True, **plotargs):
    """ plot Cumulative Ratio. """
    n_samples = len(data)
    X = sorted(data, reverse=reverse)
    Y = np.arange(1,n_samples+1)/n_samples

    if plot or ax:
        if ax is None:
            fig, ax = plt.subplots()
        ax.plot(X, Y, **plotargs)
        ax.set_ylabel("Cumulative Ratio")
        return ax

    return (X, Y)

def plot_roc_curve(labels, scores, num=50, ax=None, plot=True, **plotargs):
    flags = labels>0
    tX,tY = plot_CDF(scores[flags], plot=False)
    fX,fY = plot_CDF(scores[~flags], reverse=True, plot=False)
    plot_CDF

    def FRR2score(y):
        idx = np.abs(np.asarray(tY) - y).argmin()
        if y==tY[idx]:
            x=tX[idx]
        else:
            idx = idx-1 if y<tY[idx] else idx
            x = ((y-tY[idx])*tX[idx+1] + (tY[idx+1]-y)*tX[idx])/(tY[idx+1]-tY[idx])
        return x

    def score2FAR(x):
        if x<fX[0]:  return fY[0]
        if x>fX[-1]: return fY[-1]
        idx = np.abs(np.asarray(fX) - x).argmin()
        if x==fX[idx]:
            y=fY[idx]
        else:
            idx = idx-1 if x<fX[idx] else idx
            if fX[idx+1]==fX[idx]:
                y = fY[idx]
            else:
                y = ((x-fX[idx])*fY[idx+1] + (fX[idx+1]-x)*fY[idx])/(fX[idx+1]-fX[idx])
        return y

    FRRs = [i/N for i in range(N+1)]
    FARs = [score2FAR(FRR2score(frr)) for frr in FRRs]
    FARs[0] = 1

    if plot or ax:
        if ax is None:
            fig, ax = plt.subplots()
        ax.plot(FRRs, FARs, **plotargs)
        ax.set_xlabel("FRR(False Rejection Rate)")
        ax.set_ylabel("FAR(False Acceptance Rate)")
        ax.set_title("The relationship between FRR and FAR")
        return ax

    return (FRRs, FARs)



"""Metric-related utilities."""

import numpy as np
from scipy.sparse import issparse

from .generic_utils import handleKeyError
from .np_utils import CategoricalEncoder

def norm_vectors(*args, axis=-1, squared=True):
    if squared:
        return tuple([np.sum(np.square(arg), axis=axis) for arg in args])
    else:
        return tuple([np.sqrt(np.sum(np.square(arg), axis=axis)) for arg in args])

def normalize_vectors(*args, axis=-1):
    return tuple([arg/np.expand_dims(np.linalg.norm(arg, axis=axis), axis=axis) for arg in args])

def mean_squared_error(y_true, y_pred, sample_weight=None):
    return np.average((y_true - y_pred)**2, axis=0, weights=sample_weight).astype(float)

def root_mean_squared_error(y_true, y_pred, sample_weight=None):
    return np.sqrt(mean_squared_error(y_true, y_pred, sample_weight=sample_weight))

def silhouette_samples(X, labels, metric='euclidean'):
    encoder = CategoricalEncoder()
    labels = encoder.to_categorical(labels)
    all_dists = pairwise_euclidean_distances(X)
    cluster_mean_dists = np.asarray([np.mean(all_dists[labels==label], axis=0) for label in np.unique(labels)]).T
    silhouette_scores = np.zeros(shape=(len(X)))
    for i,(dists,cls) in enumerate(zip(cluster_mean_dists,labels)):
        intra_dists = dists[cls]
        extra_dists = np.delete(dists, cls).min()
        silhouette_scores[i] = (extra_dists-intra_dists)/max(intra_dists,extra_dists)

    return silhouette_scores

def silhouette_score(X, labels, metric='euclidean', **kwargs):
    return np.mean(silhouette_samples(X, labels, metric=metric, **kwargs))

# Paired Distances
def check_paired_array(X,Y):
    """ Arrange the two array shapes.
    @param  X: shape=(Nx,Dx)
    @param  Y: shape=(Ny,Dy)
    @return X: shape=(N,D)
    @return Y: shape=(N,D)
    """
    if X.ndim==1: X=X.reshape(1,-1)
    if Y.ndim==1: Y=Y.reshape(1,-1)

    Nx,Dx = X.shape; Ny,Dy = Y.shape

    if Dx != Dy:
        raise ValueError(f"X.shape[1], and Y.shape[1] should be the same shape, but X.shape[1]={Dx} while Y.shape[1]={Dy}")
    if Nx!=Ny:
        if Nx==1:   X=np.tile(X, [Ny, 1])
        elif Ny==1: Y=np.repeat(Y, [Nx, 1])
        else:
            raise ValueError(f"X.shape={X.shape}, and Y.shape={Y.shape}, so we couldn't understand how to pair of 2 arrays.")
    return X,Y

def paired_distances(X,Y,metric="euclidean",**kwargs):
    """Compute the paired distances between X and Y.
    @params X,Y      : shape=(N,D) should be the same.
    @param metric    : string or callable.
    @return distances: shape=(N,)
    """
    if metric in PAIRED_DISTANCES:
        func = PAIRED_DISTANCES[metric]
        return func(X,Y)
    elif callable(metric):
        X,Y = check_paired_array(X,Y)
        distances = np.asarray([metric(x,y) for x,y in zip(X,Y)])
        return distances
    else:
        handleKeyError(list(PAIRED_DISTANCES.keys()), metric=metric)

def paired_euclidean_distances(X,Y,squared=False):
    X,Y = check_paired_array(X,Y)
    if squared:
        return np.sum(np.square(X-Y), axis=1)
    else:
        return np.sqrt(np.sum(np.square(X-Y), axis=1))

def paired_manhattan_distances(X,Y):
    X,Y = check_paired_array(X,Y)
    return np.sum(np.abs(X,Y), axis=1)

def paired_cosine_distances(X,Y):
    """ ||X-Y||^2 = ||X||^2 + ||Y||^2 - 2||X||･||Y||cos(theta) """
    X,Y = check_paired_arrays(X,Y)
    X,Y = normalize_vectors(X,Y)
    return .5 * paired_euclidean_distances(X-Y, squared=True)

PAIRED_DISTANCES = {
    'cosine': paired_cosine_distances,
    'euclidean': paired_euclidean_distances,
    'l2': paired_euclidean_distances,
    'l1': paired_manhattan_distances,
    'manhattan': paired_manhattan_distances,
    'cityblock': paired_manhattan_distances,
}

def _return_float_dtype(X, Y):
    """
    1. If dtype of X and Y is float32, then dtype float32 is returned.
    2. Else dtype float is returned.
    """
    if not issparse(X) and not isinstance(X, np.ndarray):
        X = np.asarray(X)

    if Y is None:
        Y_dtype = X.dtype
    elif not issparse(Y) and not isinstance(Y, np.ndarray):
        Y = np.asarray(Y)
        Y_dtype = Y.dtype
    else:
        Y_dtype = Y.dtype

    if X.dtype == Y_dtype == np.float32:
        dtype = np.float32
    else:
        dtype = np.float

    return X, Y, dtype

# Pairwise Distances
def check_pairwise_array(X,Y=None):
    """ Arrange for the pairwise distance.
    @param X: shape=(N,D)
    """
    if Y is None:
        Y = X
    X, Y, dtype_float = _return_float_dtype(X, Y)
    X = np.array(X, dtype=dtype_float)
    Y = np.array(Y, dtype=dtype_float)

    if X.ndim==1: X=X.reshape(1,-1)
    if Y.ndim==1: Y=Y.reshape(1,-1)
    Nx,Dx = X.shape; Ny,Dy = Y.shape
    if Dx != Dy:
        raise ValueError(f"X.shape={X.shape}, and Y.shape={Y.shape}, so we couldn't understand how to pairwise of 2 arrays.")
    return X,Y

def pairwise_distances(X,Y=None,metric="euclidean",**kwargs):
    """Compute the pair-wise distances between X and Y (X).
    @params X        : shape=(Nx,D)
    @params Y        : shape=(Ny,D)
    @param metric    : string or callable.
    @return distances: shape=(Nx,Ny)
    """
    if metric in PAIRWISE_DISTANCES:
        func = PAIRWISE_DISTANCES[metric]
        return func(X,Y,**kwargs)
    elif metric in PAIRED_DISTANCES:
        func = PAIRWISE_DISTANCES[metric]
        distances = np.asarray([func(x,Y) for x in X])
        return distances
    elif callable(metric):
        X,Y = check_pairwise_array(X,Y)
        distances = np.asarray([[metric(x,y) for y in Y] for x in X])
        return distances
    else:
        handleKeyError(list(PAIRWISE_DISTANCES.keys()), metric=metric)

def pairwise_euclidean_distances(X,Y=None,squared=False):
    """Calculate euclid distance
    @param X: (ndarray) shape=(Nx,D)
    @param Y: (ndarray) shape=(Ny,D)
    @param squared   : (bool) whether distances will be squared or not.
    @return distances: Dij means the distance between Xi and Xj.
    """
    X,Y = check_pairwise_array(X,Y)
    Xnorm,Ynorm = norm_vectors(X,Y,axis=-1,squared=True)
    # ||a - b||^2 = ||a||^2 + ||b||^2 - 2 <a, b>
    distances = np.expand_dims(Xnorm, axis=1) + np.expand_dims(Ynorm, axis=0) - 2*X.dot(Y.T)
    distances = np.maximum(distances, 0.0) # for numerical errors.
    if Y is X:
        np.fill_diagonal(a=distances, val=0.)
    if not squared:
        mask = np.equal(distances, 0.) # for avoid the gradient of sqrt is infinite,
        distances += mask*1e-16 # add a small epsilon where distances == 0.0
        distances = np.sqrt(distances, out=distances)
        distances = distances * np.logical_not(mask) # Correct the epsilon added

    return distances

PAIRWISE_DISTANCES = {
    'euclidean': paired_euclidean_distances,
    'l2': paired_euclidean_distances,
}



#coding:utf-8
import sys
import time
import numpy as np
import matplotlib.pyplot as plt
from IPython.display import clear_output

from .vis_utils import measureCanvas

VALUES = {}
INITIAL_TIME = 0

def flush_progress_bar(it, max_iter, metrics={}, barname="", verbose=1, **kwargs):
    """Output the realtime progress.
    @params it      : (int) Current iteration.
    @params max_iter: (int) Maximum number of iterations.
    @params verbose : (int) -1, 0, 1, 2
        -1 = silent
        0  = only progress bar
        1  = progress bar and metrics
        2  = progress plot
    @params metric  : {(str): (float, int)}
    @params barname : (str) 
    """
    if verbose<0: 
        return
    elif verbose>1:
        flush_progress_plot(it, max_iter, metrics=metrics, **kwargs)
    else:
        global INITIAL_TIME
        if it == 0: 
            INITIAL_TIME = time.time()
        if len(barname)>0 and barname[-1]!=" ":
            barname = barname + " "
        
        it+=1
        digit = len(str(max_iter))
        rate  = it/max_iter
        n_bar = int(rate/0.05)
        bar   = ('#' * n_bar).ljust(20, '-')
        percent = f"{rate*100:.2f}"


        if verbose==1:
            metric = ", ".join([f"{k}: {v}" for  k,v in metrics.items()])
            content = f"\r{barname}{it:>0{digit}}/{max_iter} [{bar}] {percent:>6}% - {time.time()-INITIAL_TIME:.3f}s  {metric}"
        else: # verbose==0
            content = f"\r{barname}{it:>0{digit}}/{max_iter} [{bar}] {percent:>6}% - {time.time()-INITIAL_TIME:.3f}s"
        sys.stdout.write(content)

def flush_progress_plot(it, max_iter, ncols_max=3, metrics={}, **kwargs):
    """If verbose==2, this function will be called.
    """
    num_values=len(metrics)
    ncols, nrows, figsize = measureCanvas(num_values, ncols_max=ncols_max, figinches=(6,4))
    
    fig,axes = plt.subplots(nrows=nrows, ncols=ncols)
    fig.set_size_inches(*figsize)
    
    it+=1
    global VALUES
    for i, (k,v) in enumerate(metrics.items()):
        if it==1:
            VALUES[k] = [v]
        else:
            clear_output(wait=True)
            VALUES[k].append(v)
        if num_values==1:
            axes.plot(VALUES[k])
            axes.set_title(f"{it}/{max_iter} {k}: {v}")
            axes.set_xlabel("iteration")
        else:
            axes[i].plot(VALUES[k])
            axes[i].set_title(f"{it+1}/{max_iter} {k}: {v}")
            axes[i].set_xlabel("iteration")
    if it<max_iter:
        plt.pause(0.05)
        clear_output(wait=True)
    else:
        plt.show()



import os
import re
import sys
import time
import datetime
import numpy as np
from six.moves.urllib.request import urlretrieve

from collections import defaultdict

def flatten_dual(lst):
    return [e for sublist in lst for e in sublist]

_UID_PREFIXES = defaultdict(int)
def get_file(fname, origin, cache_subdir='datasets', file_hash=None):
    """Downloads a file from a URL if not already in the cache.
    ref: https://github.com/keras-team/keras/blob/7a39b6c62d43c25472b2c2476bd2a8983ae4f682/keras/utils/data_utils.py#L123

    By default the file at the url `origin` is downloaded to the
    cache_dir `~/.kerasy`, placed in the cache_subdir `datasets`,
    and given the filename `fname`. The final location of a file
    `example.txt` would therefore be `~/.kerasy/datasets/example.txt`.

    You have to make a directory `~/.kerasy` and `~./kerasy/datasets`,
    and check whether you can access these directories using `os.access("DIRECOTRY", os.W_OK)`
    If this method returns False, you have to change the ownership of them like
    ```
    $ sudo chown iwasakioshuto: ~/.kerasy
    $ sudo chown iwasakioshuto: ~/.kerasy/datasets
    ```
    """
    cache_dir = os.path.join(os.path.expanduser('~'), '.kerasy')
    datadir_base = os.path.expanduser(cache_dir)
    if not os.access(datadir_base, os.W_OK):
        datadir_base = os.path.join('/tmp', '.kerasy')
    datadir = os.path.join(datadir_base, cache_subdir)
    if not os.path.exists(datadir):
        os.makedirs(datadir)
    fpath = os.path.join(datadir, fname)

    if not os.path.exists(fpath):
        print('Downloading data from', origin)
        error_msg = 'URL fetch failure on {} : {} -- {}'
        try:
            try:
                urlretrieve(origin, fpath)
            except HTTPError as e:
                raise Exception(error_msg.format(origin, e.code, e.msg))
            except URLError as e:
                raise Exception(error_msg.format(origin, e.errno, e.reason))
        except (Exception, KeyboardInterrupt):
            if os.path.exists(fpath):
                os.remove(fpath)
            raise
    return fpath

def get_varname(var, scope_=globals()):
    for varname,val in scope_.items():
        if id(val)==id(var):
            return varname

def disp_var_globals(*varnames, head_=True, align_=True, scope_=globals()):
    """
    def func():
        a = "hoge"
        b = 123
        c = [1,"1"]
        disp_var_globals("a","b","c",scope=locals())

    func()
    >>> a: hoge
    >>> b: 123
    >>> c: [1, '1']
    """
    if head_: print(f"#=== VARIABLE INFO ===")
    digit = max([len(e) for e in varnames]) if align_ else 1
    for var in varnames:
        print(f"{var:<{digit}}: {scope_.get(var)}")

def disp_val_globals(*values, head_=True, align_=True, scope_=globals()):
    """
    def func():
        a = "hoge"
        b = 123
        c = [1,"1"]
        disp_val_globals(a,b,c,scope=locals())

    func()
    >>> a: hoge
    >>> b: 123
    >>> c: [1, '1']
    """
    if head_: print(f"#=== VARIABLE INFO ===")
    names = [get_varname(val, scope_=scope_) for val in values]
    digit = max([len(e) for e in names]) if align_ else 1
    for name,val in zip(names, values):
        print(f"{name:<{digit}}: {val}")

def get_uid(prefix=""):
    _UID_PREFIXES[prefix] += 1
    return _UID_PREFIXES[prefix]

class priColor:
    BLACK     = '\033[30m'
    RED       = '\033[31m'
    GREEN     = '\033[32m'
    YELLOW    = '\033[33m'
    BLUE      = '\033[34m'
    PURPLE    = '\033[35m'
    CYAN      = '\033[36m'
    WHITE     = '\033[37m'
    RETURN    = '\033[07m'    # 反転
    ACCENT    = '\033[01m'    # 強調
    FLASH     = '\033[05m'    # 点滅
    RED_FLASH = '\033[05;41m' # 赤背景+点滅
    END       = '\033[0m'
    @staticmethod
    def color(value, color="RED"):
        color = color.upper()
        handleKeyError(priColor.__dict__.keys(), color=color)
        return f"{priColor.__dict__[color.upper()]}{value}{priColor.END}"

def handleKeyError(lst, msg_="", **kwargs):
    k,v = kwargs.popitem()
    if v not in lst:
        lst = ', '.join([f"'{e}'" for e in lst])
        raise KeyError(f"Please chose the argment `{k}` from {lst}.\n{msg_}")

def handleTypeError(type_, message_="", **kwargs):
    k,v = kwargs.popitem()
    if not isinstance(v, type_):
        str_true_type  = re.sub(r"<class '(.*?)'>", r"\1", str(type_))
        srt_false_type = re.sub(r"<class '(.*?)'>", r"\1", str(type(v)))
        raise TypeError(f"`{k}` must be {str_true_type}, not {srt_false_type}")

def urlDecorate(url, addDate=True):
    """ Decorate URL like Wget. (Add datetime information and coloring url to blue.) """
    now = datetime.datetime.now().strftime("--%Y-%m-%d %H:%M:%S--  ") if addDate else ""
    return now + priColor.color(url, color="BLUE")

def measure_complexity(func, *args, repetitions_=10, **kwargs):
    times=0
    metrics=[]
    if "random_state" in kwargs:
        base_seed = kwargs.get("random_state")
        for i in range(repetitions_):
            kwargs["random_state"] = base_seed+i
            s = time.time()
            ret = func(*args, **kwargs)
            times += time.time()-s
            metrics.append(ret)
    else:
        for _ in range(repetitions_):
            s = time.time()
            ret = func(*args, **kwargs)
            times += time.time()-s
            metrics.append(ret)
    if metrics[0] is None:
        return times/repetitions_
    else:
        return (times/repetitions_, metrics)

def has_not_attrs(obj, *names):
    return [name for name in names if not hasattr(obj, name)]

def has_all_attrs(obj, *names):
    return sum([1 for name in names if not hasattr(obj, name)])==0

def handle_random_state(seed):
    """ Turn `np.random.RandomState` """
    if seed is None:
        return np.random.mtrand._rand
    if isinstance(seed, np.random.RandomState):
        return seed
    if isinstance(seed, int):
        return np.random.RandomState(seed)
    raise ValueError(f"Could not conver {seed} to numpy.random.RandomState instance.")

def fout_args(*args, sep="\t"):
    return sep.join([str(e) for e in args])+"\n"

f_aligns           = ["<", ">", "=", "^"]
f_signs            = ["+", "-", " ", ""]
f_grouping_options = ["_", ",", ""]
f_types            = ["b", "c", "d", "e", "E", "f", "F", "g", "G", "n", "o", "s", "x", "X", "%"]

def format_spec_create(width, align=">", sign="", zero_padding=False,
                       grouping_option="", fmt=""):
    """
    Create a function which returns a formatted text.
    ~~~~~
    * Source Code  : https://github.com/python/cpython/blob/3.8/Lib/string.py
    * Documentation:  https://docs.python.org/3/library/string.html#format-specification-mini-language

    format_spec = [[fill]align][sign][#][0][width][grouping_option][.precision][type]
    =========================
    @params align           : [[fill]align]
    @params sign            : [sign]
    @params zero_padding    : [0]
    @params width           : [width]
    @params grouping_option : [grouping_option]
    @params fmt             : [.precision][type]
    @return lambda          : <function __main__.<lambda>(fill)>
    """
    handleKeyError(lst=f_aligns, align=align)
    handleKeyError(lst=f_signs,  sign=sign)
    handleKeyError(lst=f_grouping_options, grouping_option=grouping_option)
    if len(fmt)>0:
        handleKeyError(lst=f_types, fmt=fmt[-1])
    zero = "0" if zero_padding else ""
    handleTypeError(type_=int, width=width)
    return lambda fill : f"{fill:{align}{sign}{zero}{width}{grouping_option}{fmt}}"

def print_func_create(width, align=">", sign="", zero_padding=False,
                      grouping_option="", fmt="", color="black",
                      left_side_bar="", right_side_bar="",
                      left_margin=0, right_margin=0, end="\n"):
    """
    Create a function which prints a formatted text.
    Please see also the function `format_spec_create`.
    ==============================
    @params color                : string color
    @params left(right)_side_bar : (str)
    @params left(right)_margin   : (int)
    @params end                  : string appended after the last value, default a newline.
    @return lambda               : <function __main__.<lambda>(fill)>
    """
    format_spec = format_spec_create(width, align=align, sign=sign,
                                     zero_padding=zero_padding,
                                     grouping_option=grouping_option, fmt=fmt)
    def print_func(fill):
        info  = f"{left_side_bar}{' '*left_margin}"
        info += priColor.color(format_spec(fill), color=color)
        info += f"{' '*right_margin}{right_side_bar}"
        print(info, end=end)
    return print_func

class Table():
    def __init__(self):
        self.cols = {}
        self.table_width = 1
        self.head = None

    def _disp_title(self):
        for colname, options in self.cols.items():
            if "print_values" not in options:
                continue
            print_func = options.get("print_title")
            print_func(colname)
        print("|")

    def _disp_border(self, table_width=None, mark="="):
        table_width = self.table_width if table_width is None else table_width
        print(mark*table_width)

    def _disp_values(self, head=None):
        head = self.head if head is None else head
        for i in range(head):
            for colname, options in self.cols.items():
                if "print_values" not in options:
                    continue
                print_func = options.get("print_values")
                values = options.get("values")
                print_func(values[i])
            print("|")

    def show(self, head=None, table_width=None, mark="="):
        self._disp_title()
        self._disp_border(table_width=table_width, mark=mark)
        self._disp_values(head=head)

    def set_cols(self, colname, values, width=None, align=">", sign="",
                 zero_padding=False, grouping_option="", fmt="", color="black",
                 left_margin=0, right_margin=0):
        title_width = len(str(colname))
        if width is None:
            format_spec = format_spec_create(
                width=0, align=align, sign=sign, zero_padding=zero_padding,
                grouping_option=grouping_option, fmt=fmt
            )
            width = len(max([format_spec(v) for v in values], key=len))
        width = max(width, title_width)
        self.table_width += width + left_margin + right_margin + 1

        print_values = print_func_create(
            width=width, align=align, sign=sign, zero_padding=zero_padding,
            grouping_option=grouping_option, fmt=fmt, color=color,
            left_side_bar="|", right_side_bar="", end="",
            left_margin=left_margin, right_margin=right_margin,
        )
        print_title = print_func_create(
            width=width, align="^", sign="", zero_padding=False,
            grouping_option="", fmt="", color="ACCENT",
            left_side_bar="|", right_side_bar="", end="",
            left_margin=left_margin, right_margin=right_margin,
        )
        self.cols.update({colname: dict(
            print_values=print_values, print_title=print_title, values=values
        )})
        if self.head is None:
            self.head = len(values)



# coding: utf-8
from __future__ import absolute_import

import numpy as np
from collections import defaultdict

def clip_norm(grads,c,n):
    """Ref: https://github.com/keras-team/keras/blob/7a39b6c62d43c25472b2c2476bd2a8983ae4f682/keras/optimizers.py#L21
    Clip the gradient `g` if the L2 norm `n` exceeds `c`.
    =====================================================
    @param  g: (ndarray) Gradient.
    @param  c: (float)   Gradients will be clipped when their L2 norm exceeds this value.
    @param  n: (ndarray) Actual norm of `g`
    @return g: (ndarray) Gradient clipped if required.
    """
    if c<=0: return g
    g = c/n*g if n>=c else g
    return g

class Optimizer():
    """Abstract optimizer base class.
    ALL Kerasy optimizer support the following keyword arguments:
    @param clipnorm : (float >= 0) Gradients will be clipped when their L2 norm exceeds this value.
    @param clipvalue: (float >= 0) Gradients will be clipped when their absolute value exceeds this value.
    """
    def __init__(self, **kwargs):
        allowed_kwargs = {"clipnorm", "clipvalue"}
        for k in kwargs:
            if k not in allowed_kwargs:
                raise TypeError(f'Unexpecte keyword argument passed to optimizer: {str(k)}')

        self.__dict__.update(kwargs)
        self.updates = []
        self.weights = []
        self.optparams = defaultdict(lambda: defaultdict(list))

    def get_updates(self, grads, past_params, name):
        """ This is the parent class of all optimizer, not an actual optimizer. """
        raise NotImplementedError

    def get_gradients(self, grads):
        if hasattr(self, 'clipnorm') and self.clipnorm > 0:
            norm = np.sqrt(sum([np.sum(np.square(g)) for g in grads]))
            grads = [clip_norm(g, self.clipnorm, norm) for g in grads]
        if hasattr(self, 'clipvalue') and self.clipvalue > 0:
            grads = [np.clip(g, -self.clipvalue, self.clipvalue) for g in grads]
        return grads

class GradientDescent(Optimizer):
    def __init__(self, eta=0.01, **kwargs):
        self.eta = eta
        super().__init__(**kwargs)

    def get_updates(self, grads, past_params, name):
        grads = self.get_gradients(grads) # Cliped.
        last_param = past_params[-1]
        return last_param - self.eta*grads

class SGD(Optimizer):
    """Stochastic gradient descent optimizer.
    @param learning_rate: (float) Learning rate.
    @param momentum     : (float) Parameter that accelerates SGD in the relevant direction and dampens oscillations.
    @param nesterov     : (bool)  Whether to apply Nesterov momentum.
    """
    def __init__(self, learning_rate=0.01, momentum=0.9, nesterov=False, **kwargs):
        learning_rate = kwargs.pop('lr', learning_rate) # dict.pop(key,d): If key is not found, d is returned
        self.initial_decay = kwargs.pop('decay', 0.0)
        super().__init__(**kwargs)
        self.iterations = 0
        self.learning_rate = learning_rate
        self.momentum = momentum
        self.decay = self.initial_decay
        self.nesterov = nesterov

    def get_updates(self, grads, past_params, name):
        """
        @grads      : (ndarray) shape=(x,y)
        @past_params: (ndarray) shape=(z,x,y)
        @name       : (str) layer name + param name.
        """
        grads = self.get_gradients(grads) # Cliped.
        # Increase processing speed by storing in local variables.
        if self.initial_decay > 0:
            lr *= (1. / (1. + self.decay*self.iterations))
        if past_params.shape[0] <= 1:
            dParam = np.zeros_like(grads)
            Last = past_params[-1]
        else:
            preLast, Last = past_params[-2:]
            dParam = Last-preLast
        velocity = self.momentum*dParam - self.learning_rate*grads
        if self.nesterov:
            new_param = Last + self.momentum*velocity - self.lr*grads
        else:
            new_param = Last + velocity
        return new_param

OptimizerHandler = {
    'sgd': SGD,
    'gra': GradientDescent,
}

def Optimizer(optimizer_name):
    return OptimizerHandler[optimizer_name]



# coding: utf-8
from __future__ import absolute_import

import numpy as np

from ..activations import ActivationFunc
from ..initializers import Initializer
from ..losses import LossFunc
from ..engine.base_layer import Layer

class Conv2D(Layer):
    def __init__(self, filters, kernel_size=(3,3), strides=(1,1), padding='valid', activation='relu',
                 kernel_initializer='random_normal', bias_initializer='zeros', data_format='channels_last', **kwargs):
        """
        @param filters           : (int) the dimensionality of the output space.
        @param kernel_size       : (int,int) height and width of each kernel. kernel-shape=(*kernel_size, F, OF)
        @param strides           : (int,int) height and width of stride step.
        @param padding           : (str) "valid" or "same". Padding method.
        @param activation        : (str) Activation function to use.
        @param kernel_initializer: (str) Initializer for the `kernel` weights matrix.
        @param bias_initializer  : (str) Initializer for the bias vector.
        """
        self.OF = filters # Output filters.
        self.kh, self.kw = kernel_size # kernel size.
        self.sh, self.sw = strides
        if padding not in ["same", "valid"]: raise ValueError("padding must be 'same' or 'valid'. Please chose one of them.")
        self.padding = padding
        self.h = ActivationFunc(activation)
        self.kernel_initializer = Initializer(kernel_initializer)
        self.kernel_regularizer = None
        self.kernel_constraint  = None
        self.bias_initializer   = Initializer(bias_initializer)
        self.bias_regularizer   = None
        self.bias_constraint    = None
        self.use_bias = True
        super().__init__(**kwargs)

    def compute_output_shape(self, input_shape):
        if len(input_shape) != 3: raise ValueError(f"The input shape of {self.name} must be 3-dimension (height, width, channel). However it is {len(input_shape)}-dimension.")
        self.H, self.W, self.F = input_shape
        if self.padding=="same":
            self.OH = self.H
            self.OW = self.W
            self.ph = ((self.sh-1)*self.H+self.kh-self.sh)//2
            self.pw = ((self.sw-1)*self.W+self.kw-self.sw)//2
        elif self.padding=="valid":
            self.OH = (self.H-self.kh)//self.sh+1
            self.OW = (self.W-self.kw)//self.sw+1
            self.ph = 0
            self.pw = 0
        self.output_shape = (self.OH, self.OW, self.OF)
        return self.output_shape

    def build(self, input_shape):
        """ @params input_shape: (H,W,F) of input image. """
        output_shape = self.compute_output_shape(input_shape)
        self.kernel  = self.add_weight(shape=(self.kh, self.kw, self.F, self.OF),
                                       name="kernel",
                                       initializer=self.kernel_initializer,
                                       regularizer=self.kernel_regularizer,
                                       constraint =self.kernel_constraint,
                                       trainable  =self.trainable)
        if self.use_bias:
            self.bias = self.add_weight(shape=(self.OF,),
                                        name="bias",
                                        initializer=self.bias_initializer,
                                        regularizer=self.bias_regularizer,
                                        constraint =self.bias_constraint,
                                        trainable  =self.trainable)
        else:
            self.bias = None
        return output_shape

    def _paddInput(self, input):
        """ @param input: (ndarray) must be a 3-D array. shape=(H,W,F) """
        #=== padding ===
        if self.padding =="same":
            Xin = np.zeros(shape=(self.H+2*self.ph,self.W+2*self.pw,self.F))
            Xin[self.ph:self.H+self.ph,self.pw:self.W+self.pw,:] = input
        elif self.padding == "valid":
            Xin = input[:self.sh*self.OH+self.kh-1, :self.sw*self.OW+self.kw-1, :]
        else:
            TypeError(f"Can't understand 'padding=f{self.padding}'")
        self.Xin = Xin # Memorize
        return Xin

    def _generator(self, Xin):
        #=== generator ===
        for i in range(self.OH//self.sh):
            for j in range(self.OW//self.sw):
                clipedXin = Xin[self.sh*i:(self.sh*i+self.kh), self.sw*j:(self.sw*j+self.kw), :, None]
                yield clipedXin,i,j

    # old version: https://github.com/iwasakishuto/Kerasy/blob/c6a896834be7703e0454ba44ffcd8a66e5de197c/kerasy/layers/convolutional.py#L94
    def forward(self, input):
        """ @param input: (ndarray) 3-D array. shape=(H,W,F) """
        Xin  = self._paddInput(input)
        a    = np.empty(shape=(self.OH, self.OW, self.OF))
        for clip_image,i,j in self._generator(Xin):
            """ 'self.kernel' and np.expand_dims('clip_image', axis=-1) shapes equal to ( kh,kw,F,OF(None) ) """
            a[i,j,:] = np.sum(clip_image*self.kernel, axis=(0,1,2))
        a += self.bias # (OH,OW,OF) + (OF,) = (OH,OW,OF)
        self.a = a     # Memorize. (output layer. shape=(OH,OW,OF))
        Xout = self.h.forward(a)
        return Xout

    def _backprop_mask(self,i,j,m,n):
        return ((i%self.sh+m < self.kh) and (j%self.sw+n < self.kw)) and ((self.OH > (i-m)//self.sh >= 0) and (self.OW > (j-n)//self.sw >= 0))

    def backprop(self, dEdXout, lr=1e-3):
        """
        @param  delta_times_w: shape=(OH,OW,OF)
        @param  self.kernel  : shape=(self.kh, self.kw, self.F, self.OF)
        @return delta_times_w: shape=(H,W,F)
        """
        dEda = dEdXout*self.h.diff(self.a) # Xout=h(a) → dE/da = dE/dXout*h'(a)
        dEdXin = np.zeros_like(self.Xin)   # shape=(H+2ph,W+2pw,F)
        for c in range(self.F):
            for i in range(self.H+2*self.ph):
                for j in range(self.W+2*self.pw):
                    dEdXin[i,j,c] = np.sum([ dEda[(i-m)//self.sh,(j-n)//self.sw,:] * self.kernel[i%self.sh+m,j%self.sw+n,c,:] for m in range(0,self.kh,self.sh) for n in range(0,self.kw,self.sw) if self._backprop_mask(i,j,m,n)])
        if self.trainable: self.memorize_delta(dEda)
        return dEdXin[self.ph:self.H+self.ph,self.pw:self.W+self.pw,:]

    def memorize_delta(self, dEda):
        dEdw = np.zeros(shape=self.kernel.shape) # shape=(kh, kw, F, OF)
        for m in range(self.kh):
            for n in range(self.kw):
                for c in range(self.F):
                    for c_ in range(self.OF):
                        # dEdw_{m,n,c,c'} = ΣiΣj(dEda * dadw) = ΣiΣj(dEda_{i,j,c'}*Xin_{i+m,j+n,c})
                        dEdw[m,n,c,c_] = np.sum(dEda[:,:,c_] * self.Xin[m:m+self.OH:self.sh, n:n+self.OW:self.sw, c])
        self._losses['kernel'] += dEdw
        self._losses['bias'] += np.sum(dEda, axis=(0,1))



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import numpy as np

from ..losses import LossFunc
from ..optimizers import Optimizer
from ..layers.core import Input
from .base_layer import Layer
from ..utils import make_batches
from ..utils import flush_progress_bar
from ..utils import print_summary

class Sequential():
    def __init__(self):
        self.layers = []

    def add(self, layer):
        """Adds a layer instance."""
        if not isinstance(layer, Layer):
            raise TypeError(f"The added layer must be an instance of class Layer. Found: {str(layer)}")
        self.layers.append(layer)

    def compile(self, optimizer, loss=None, metrics=None):
        """ Creates the layer weights.
        @param optimizer: (String name of optimizer) or (Optimizer instance).
        @param loss     : (String name of loss function) or (Loss instance).
        @param metrics  : (List) Metrics to be evaluated by the model during training and testing.
        """
        self.optimizer = Optimizer(optimizer)() if isinstance(optimizer, str) else optimizer
        self.loss = LossFunc(loss) if isinstance(loss, str) else loss
        self.metrics = metrics
        input_layer = self.layers[0]
        if not isinstance(input_layer, Input): raise ValueError(f"The initial layer should be Input instance, but {str(input_layer)}")
        output_shape = input_layer.input_shape
        for layer in self.layers:
            output_shape = layer.build(output_shape)

    def fit(self,
            x=None, y=None, batch_size=32, epochs=1, verbose=1, shuffle=True,
            validation_spilit=0, validation_data=None, validation_steps=None,
            class_weight=None, sample_weight=None, **kwargs):
        if kwargs: raise TypeError(f'Unrecognized keyword arguments: {str(kwargs)}')
        if (x is None) or (y is None): raise ValueError('Please specify the trainig data. (x,y)')
        # Prepare validation data.
        do_validation = False
        if validation_data:
            do_validation = True
            if len(validation_data) == 2:
                val_x, val_y = validation_data
                val_sample_weight = None
            elif len(validation_data) == 3:
                val_x, val_y, val_sample_weight = validation_data
            else: raise ValueError(f"When passing validation_data, it must contain 2 (x_val, y_val) or 3 (x_val, y_val, val_sample_weights) items. However, it contains {len(validation)} items.")

        # Prepare for the trainig.
        epoch_digit = len(str(epochs))
        num_train_samples = len(x)
        batches = make_batches(num_train_samples, batch_size)
        num_batchs = len(batches)
        index_array = np.arange(num_train_samples)

        for epoch in range(epochs):
            if shuffle: np.random.shuffle(index_array)
            losses = 0
            for batch_index, (batch_start, batch_end) in enumerate(batches):
                batch_ids = index_array[batch_start:batch_end]
                for bs, (x_train, y_true) in enumerate(zip(x[batch_ids], y[batch_ids])):
                    y_pred = self.forward(x_train)
                    losses += self.loss.loss(y_true=y_true, y_pred=y_pred)
                    self.backprop(y_true=y_true, y_pred=y_pred)
                self.updates(bs+1)
                flush_progress_bar(batch_index,
                                   num_batchs,
                                   barname=f"Epoch {epoch+1:>0{epoch_digit}}/{epochs} |",
                                   metrics=f"{self.loss.__name__}: {losses/min((batch_index+1)*batch_size, num_train_samples):.4f}",
                                   verbose=verbose)
            if verbose>=1: print()

    def forward(self, input):
        out=input
        for layer in self.layers:
            out = layer.forward(out)
        return out

    def backprop(self, y_true, y_pred):
        dEdXout = self.loss.diff(y_true, y_pred)
        for layer in reversed(self.layers):
            dEdXout = layer.backprop(dEdXout)

    def predict(self, x_train):
        if np.ndim(x_train) == 1:
            return self.forward(x_train)
        else:
            return np.array([self.forward(x) for x in x_train])

    def updates(self, batch_size):
        for layer in reversed(self.layers):
            layer.update(self.optimizer, batch_size)

    def summary(self):
        print_summary(self)



# coding: utf-8
import numpy as np

class Linear():
    def forward(input):
        return input

    def diff(delta):
        return delta

class Softmax():
    def forward(input):
        """ @param input shape=(class,) """
        exps = np.exp(input - np.max(input)) # For avoiding overflowing.
        return exps/np.sum(exps)

    def diff(delta):
        """
        @param  : shape=(class,)
        @return : shape=(class,class)
        =================================
        ∂z_i    / z_i(1-z_i)  --- if i==k
        ---- = |
        ∂a_k    \ -z_iz_k     --- if i!=k
        """
        exps = np.exp(delta - np.max(delta)) # For avoiding overflowing.
        y = exps/np.sum(exps)
        return np.diag(y) - np.outer(y,y)

class Tanh():
    def forward(input):
        return np.tanh(input)

    def diff(delta):
        return 1-np.tanh(delta)**2

class Relu():
    def forward(input):
        return np.where(input>0, input, 0)

    def diff(delta):
        return np.where(delta>0, 1, 0)

class Sigmoid():
    def __init__(self):
        self.out = None

    def forward(self, input):
        out = 1/(1+np.exp(input))
        self.out = out # Memorize!!
        return out

    def diff(self, delta):
        return delta * (1-self.out) * self.out

ActivationHandler = {
    'linear' : Linear,
    'softmax': Softmax,
    'tanh'   : Tanh,
    'relu'   : Relu,
    'sigmoid': Sigmoid(),
}

def ActivationFunc(activation_func_name):
    return ActivationHandler[activation_func_name]



# coding: utf-8
from ..utils import flush_progress_bar

def breakdown(combs):
    """ display breakdowns """
    use_coins = sorted(set(combs))
    num_coins = [combs.count(coin) for coin in use_coins]
    total_pay = [n*coin for n,coin in zip(use_coins,num_coins)]
    width_coin  = max([len(str(e)) for e in use_coins]+[len("coins")])
    width_num   = max([len(str(e)) for e in num_coins]+[len("number")])
    width_total = max([len(str(e)) for e in total_pay]+[len("pay"),len(str(sum(total_pay)))])
    width_line  = width_coin+width_num+width_total+2

    print_func = lambda c,n,p: print(f"{c:^{width_coin}}|{n:>{width_num}}|{p:>{width_total}}")
    print_func('coins','number','pay')
    print("="*width_line)
    for coin,num,t in zip(use_coins,num_coins,total_pay):
        print_func(coin,num,t)
    print("-"*width_line)
    print_func('total',sum(num_coins),sum(total_pay))

def smart_pay(coins, total, limit=None, verbose=1):
    """
    Find the minimum number of coin combinations by using Dynamic Programming.
    @params coins: (int list) Coins.
    @params total: (int) Amount of Payment.
    @params limit: (int) Maximum number of times a restricted coin can be used.
    """
    total += 1 # because 0-origin.
    if len(set(coins)) < len(coins):
        raise ValueError("All elements of `coins` must be different integers.")
    restricted = coins[0]
    free_coins = coins[1:]

    if limit is None:
        limit = total//restricted+1
    elif verbose:
        print(f'{restricted} coin can only be used up to {limit} times at the same time.')

    # Initialization.
    B = [0 for _ in range(total)] # Memory for Traceback.
    m = [0 if t==0 else 1 if t in free_coins else float('inf') for t in range(total)]

    # Recursion
    for t in range(1,total):
        cands = [m[t-coin] if (t-coin)>=0 else float('inf') for coin in free_coins]
        if not sum([e!=float('inf') for e in cands])==0:
            minnum = min(cands)
            m[t],B[t] = [(e+1,t-coin) for e,coin in zip(cands,free_coins) if e==minnum][0]
        flush_progress_bar(t-1, total-1, metrics={"minimum": m[t]}, verbose=verbose)

    ms = [(l,m[-1-restricted*l]+l) for l in range(limit+1) if restricted*l<=total]
    num_restricted, num_total = min(ms, key=lambda x:x[1])
    idx = total-1-restricted*num_restricted
    combs = [restricted for _ in range(num_restricted)]
    while idx:
        last = B[idx]
        combs.append(idx-last)
        idx = last
    if verbose: breakdown(combs)



#coding: utf-8
import sys
from collections import deque
from .progress_utils import flush_progress_bar

class BaseMonitor():
    """ Monitors and reports parameters. """
    def __init__(self, n_target=1, max_iter=300, verbose=1, maxlen=2):
        self.iter = 0
        self.max_iter = max_iter
        self.histories = [deque(maxlen=maxlen) for _ in range(n_target)]
        self.verbose = verbose

    def __repr__(self):
        return f"{super().__repr__()}\n" + \
                "\n".join([f"  - {var} = {val}" for var,val in sorted(self.__dict__.items())])

    def _reset(self):
        self.iter = 0
        for hist in self.histories:
            hist.clear()

    def report(self, *targets, metrics={}, barname="", **kwargs):
        flush_progress_bar(self.iter, self.max_iter, metrics=metrics, barname=barname, verbose=self.verbose, **kwargs)
        for i,target in enumerate(targets):
            self.histories[i].append(target)
        self.iter += 1

class ConvergenceMonitor(BaseMonitor):
    def __init__(self, *tols, max_iter=300, verbose=1):
        self.tols = tols
        super().__init__(max_iter=max_iter, n_target=len(tols), verbose=verbose, maxlen=2)

    @property
    def converged(self):
        return self.iter == self.max_iter or \
               (len(self.histories[0])==2 and all([prev-current < tol for (current, prev), tol in zip(self.histories, self.tols)]))


class ThresholdMonitor(BaseMonitor):
    def __init__(self, *thresholds, max_iter=300, verbose=1):
        """
        @params threshold : (thre, mode) mode should be chosed by 'u'(upper), or 'l'(lower)
        """
        self.thresholds = thresholds
        super().__init__(max_iter=max_iter, n_target=len(thresholds), verbose=verbose, maxlen=1)

    @property
    def converged(self):
        return self.iter == self.max_iter or \
               all([val <= tol if mode[0]=='u' else val >= tol for val, (thre, mode) in zip(self.histories, self.thresholds)])



# coding: utf-8
import numpy as np

class mean_squared_error():
    def loss(y_true, y_pred):
        return np.mean(np.square(y_pred - y_true), axis=-1)
    def diff(y_true, y_pred):
        return y_pred-y_true

class categorical_crossentropy():
    def loss(y_true, y_pred):
        small_val = 0 #1e-7 # Avoiding np.log(y_pred)=inf if y_pred==0
        return -np.sum(y_true * np.log(y_pred+small_val))
    def diff(y_true, y_pred):
        small_val = 0 #1e-7
        return -y_true/(y_pred+small_val)

LossHandler = {
    'mean_squared_error' : mean_squared_error,
    'categorical_crossentropy' : categorical_crossentropy,
}

def LossFunc(loss_func_name):
    if loss_func_name not in LossHandler:
        raise KeyError(f"Please select loss function from {list(LossHandler.keys())}")
    return LossHandler[loss_func_name]



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import numpy as np

from ..utils.generic_utils import get_uid
from ..initializers import Initializer

class Layer():
    """Abstract base layer class."""
    def __init__(self, **kwargs):
        self._trainable_weights = []
        self._non_trainable_weights = []
        self._losses = {}  # (name, delta)
        self._updates = {}
        prefix = self.__class__.__name__.lower()
        self.name = prefix + '_' + str(get_uid(prefix))
        self.trainable = kwargs.get('trainable', True)

    def compute_output_shape(self, input_shape):
        """Computes the output shape of the layer."""
        output_shape = input_shape
        self.output_shape = output_shape
        return output_shape

    def build(self, input_shape):
        output_shape = self.compute_output_shape(input_shape)
        return output_shape

    def add_weight(self, shape=None, name=None, dtype=None, initializer=None, regularizer=None, constraint=None, trainable=True):
        """
        @param  shape      : (tuple) The shape of the weight.
        @param  dtype      : (dtype) The dtype of the weight.
        @param  initializer: (string) An Initializer instance.
        @param  regularizer: (string) A Regularizer instance.
        @param  trainable  : (bool) A boolean, whether the weight should be trained via backprop or not.
        @return weight     : (ndarray) The created weights variable.
        """
        shape = () if shape is None else shape
        weight = initializer(shape=shape, dtype=dtype)
        if trainable:
            self._trainable_weights.append(name)
        else:
            self._non_trainable_weights.append(name)
        self._updates[name] = np.expand_dims(weight, axis=0) # shape=(z,x,y)
        self._losses[name] = np.zeros_like(weight) # shape=(x,y)
        return weight

    def update(self, optimizer, batch_size):
        if self.trainable and len(self._non_trainable_weights)>0:
            self._trainable_weights += self._non_trainable_weights
            self._non_trainable_weights = []
        elif self.trainable == False and len(self._trainable_weights)>0:
            self._non_trainable_weights += self._trainable_weights
            self._trainable_weights = []

        for name in self._trainable_weights:
            new_weight = optimizer.get_updates(self._losses[name]/batch_size, self._updates[name], self.name)
            self.__dict__[name] = new_weight
            self._updates[name] = np.r_[self._updates[name], np.expand_dims(new_weight, axis=0)]
            self._losses[name]  = np.zeros_like(new_weight)



# coding: utf-8
import numpy as np
from scipy import stats

def Zeros(shape, dtype=None):
    return np.zeros(shape=shape, dtype=dtype)

def Ones(shape, dtype=None):
    return np.ones(shape=shape, dtype=dtype)

def Constant(shape, value=0, dtype=None):
    return np.full(shape=shape, fill_value=value, dtype=dtype)

def RandomNormal(shape, mean=0, stddev=0.05, dtype=None, seed=None):
    rng = np.random.RandomState(seed) if seed is not None else np.random
    return rng.normal(size=shape, loc=mean, scale=stddev).astype(dtype)

def RandomUniform(shape, minval=-0.05, maxval=0.05, dtype=None, seed=None):
    rng = np.random.RandomState(seed) if seed is not None else np.random
    return rng.uniform(size=shape, low=minval, high=maxval)

def TruncatedNormal(shape, mean=0.0, stddev=0.05, dtype=None, seed=None):
    X = stats.truncnorm(
        (-stddev - mean) / stddev,
        (stddev  - mean) / stddev,
        loc=mean,
        scale=stddev,
    )
    return X.rvs(size=shape,random_state=seed).astype(dtype)

def VarianceScaling(shape, scale=1.0, mode='fan_in', distribution='normal', dtype=None, seed=None):
    fan_in, fan_out = _compute_fans(shape)
    n = fan_in if mode=="fan_in" else fan_out if mode=="fan_out" else (fan_in+fan_out)/2
    scale /= max(1., n)
    if distribution=='normal':
        # 0.879... = scipy.stats.truncnorm.std(a=-2, b=2, loc=0., scale=1.)
        stddev = np.sqrt(scale) / .87962566103423978
        return TruncatedNormal(shape=shape, mean=0.0, stddev=stddev, dtype=dtype, seed=seed)
    else:
        limit = np.sqrt(3 * scale)
        return RandomUniform(shape=shape, minval=-limit, maxval=limit, dtype=dtype, seed=seed)

def Orthogonal(shape, gain=1.0, dtype=None, seed=None):
    num_rows = 1
    for dim in shape[:-1]:
        num_rows *= dim
    num_cols = shape[-1]
    flat_shape = (num_rows, num_cols)
    rng = np.random
    if seed is not None:
        rng = np.random.RandomState(seed)
    a = rng.normal(loc=0.0, scale=1.0, size=flat_shape).astype(dtype)
    u, _, v = np.linalg.svd(a, full_matrices=False)
    # Pick the one with the correct shape.
    q = u if u.shape == flat_shape else v
    q = q.reshape(shape)
    return gain * q[:shape[0], :shape[1]]

def Identity(shape, dtype=None, gain=1.0):
    if len(shape) != 2 or shape[0]!=shape[1]:
        raise ValueError('Identity matrix initializer can only be used for 2D Square matrices.')
    return gain * np.eye(N=shape[0], dtype=dtype)

def GlorotNormal(shape, dtype=None, seed=None):
    return VarianceScaling(
        shape=shape,
        scale=1.,
        mode='fan_avg',
        distribution='normal',
        dtype=dtype,
        seed=seed
    )

def GlorotUniform(shape, dtype=None, seed=None):
    return VarianceScaling(
        shape=shape,
        scale=1.,
        mode='fan_avg',
        distribution='uniform',
        dtype=dtype,
        seed=seed
    )

def HeNormal(shape, dtype=None, seed=None):
    return VarianceScaling(
        shape=shape,
        scale=2.,
        mode='fan_in',
        distribution='normal',
        dtype=dtype,
        seed=seed
    )

def LeCunNormal(shape, dtype=None, seed=None):
    return VarianceScaling(
        shape=shape,
        scale=1.,
        mode='fan_in',
        distribution='normal',
        dtype=dtype,
        seed=seed
    )

def HeUniform(shape, dtype=None, seed=None):
    return VarianceScaling(
        shape=shape,
        scale=2.,
        mode='fan_in',
        distribution='uniform',
        dtype=dtype,
        seed=seed
    )

def LeCunUniform(shape, dtype=None, seed=None):
    return VarianceScaling(
        shape=shape,
        scale=1.,
        mode='fan_in',
        distribution='uniform',
        dtype=dtype,
        seed=seed
    )

def _compute_fans(shape, data_format='channels_last'):
    """Computes the number of input and output units for a weight shape.
    @param shape      : Integer shape tuple.
    @param data_format: Image data format to use for convolution kernels.
    @return fan_in    : size of the input shape.
    @return fan_out   : size of the output shape.
    """
    if len(shape) == 2:
        fan_in,fan_out = shape
    elif len(shape) in {3, 4, 5}:
        # Assuming convolution kernels (1D, 2D or 3D).
        # TH kernel shape: (depth, input_depth, ...)
        # TF kernel shape: (..., input_depth, depth)
        if data_format == 'channels_first':
            receptive_field_size = np.prod(shape[2:])
            fan_in = shape[1] * receptive_field_size
            fan_out = shape[0] * receptive_field_size
        elif data_format == 'channels_last':
            receptive_field_size = np.prod(shape[:-2])
            fan_in = shape[-2] * receptive_field_size
            fan_out = shape[-1] * receptive_field_size
        else:
            raise ValueError('Invalid data_format: ' + data_format)
    else:
        # No specific assumptions.
        fan_in = np.sqrt(np.prod(shape))
        fan_out = np.sqrt(np.prod(shape))
    return fan_in, fan_out


InitializeHandler = {
    'zeros': Zeros,
    'ones': Ones,
    'constant': Constant,
    'random_normal': RandomNormal,
    'random_uniform': RandomUniform,
    'truncated_normal': TruncatedNormal,
    'variance_scaling': VarianceScaling,
    'orthogonal': Orthogonal,
    'identity': Identity,
    'glorot_normal': GlorotNormal,
    'glorot_uniform': GlorotUniform,
    'he_normal': HeNormal,
    'lecun_normal': LeCunNormal,
    'he_uniform': HeUniform,
    'lecun_uniform': LeCunUniform,
}

def Initializer(initializer_name):
    if initializer_name not in InitializeHandler:
        raise KeyError(f"Please select initializer from {list(InitializeHandler.keys())}")
    return InitializeHandler[initializer_name]



"""Numpy-related utilities."""
import numpy as np
from scipy import linalg
from scipy.special import logsumexp

from .generic_utils import handleKeyError

class CategoricalEncoder():
    def __init__(self):
        self.obj2cls = {}
        self.cls2obj = {}
        self.dtype = None

    def _mk_obj2cls_dict(self, obj, origin=0):
        if len(self.obj2cls) == 0:
            arr = np.asarray(list(obj))
            unique = np.unique(arr)
            self.obj2cls = dict(zip(unique, range(origin, origin+len(unique))))
            self.cls2obj = dict(zip(range(origin, origin+len(unique)), unique))
            self.dtype = arr.dtype
        else:
            print("Dictionaly for Encoder is already made.")

    def to_categorical(self, obj, origin=0):
        self._mk_obj2cls_dict(obj, origin=origin)
        return np.asarray([self.obj2cls[e] for e in obj], dtype=int)

    def to_onehot(self, obj, num_classes=None, dtype='float32', origin=0):
        n = len(obj)
        self._mk_obj2cls_dict(obj, origin=origin)
        num_classes = num_classes if num_classes is not None else len(self.obj2cls)
        categorical = np.zeros(shape=(n, num_classes), dtype=int)
        categorical[np.arange(n), np.asarray([self.obj2cls[e] for e in obj], dtype=int)] = 1
        return categorical

    def reverse(self, cls):
        return np.asarray([self.cls2obj[c] for c in cls], dtype=self.dtype)

def findLowerUpper(data, margin=1e-2, N=None):
    """ data.shape=(N,D) """
    mins = np.min(data, axis=0)
    maxs = np.max(data, axis=0)
    margins = (maxs-mins)*margin
    lbs = mins-margins; ubs = maxs+margins
    if N is None:
        return (lbs, ubs)
    else:
        return np.concatenate([[np.linspace(lb,ub,N) for lb,ub in zip(lbs, ubs)]])

def inverse_arr(arr):
    """
    arr[k]=v → arr_inv[k]=v
    """
    arr_inv = np.zeros_like(arr)
    for i in range(len(arr)):
        arr_inv[arr[i]] = i
    return arr_inv

def _check_sample_weight(sample_weight, X, dtype=np.float64):
    """ Validate sample weights.
    @params sample_weight : ndarray shape=(n_samples,) or None
    @params X             : Input data. shape=(n_samples, *n_features)
    @params dtype         : sample weight's data type.
    @return sample_weight : shape=(n_samples,)
    """
    n_samples = X.shape[0]

    if sample_weight is None:
        sample_weight = np.ones(n_samples, dtype=dtype)
    else:
        sample_weight = np.asarray(sample_weight, dtype=dtype)
        if sample_weight.ndim != 1:
            raise ValueError("Sample weights must be 1D array or scalar")
        if sample_weight.shape != (n_samples,):
            raise ValueError(f"sample_weight.shape == {sample_weight.shape}, expected {(n_samples,)}!")
    return sample_weight

def normalize(arr, axis=None):
    """ Inplace Method
    Normalizes the input array so that it sums to 1 along with `axis`.
    @params arr : Non-normalized Input array.
    @params axis: Dimension along which normalization is performed.
    """
    if axis is None:
        arr /= arr.sum()
    else:
        arr /= np.expand_dims(arr.sum(axis), axis)

def standardize(arr, axis=None):
    """ Inplace Method
    Standarzes the input array so that it sums to 1 along with `axis`.
    @params arr : Non-standarzed Input array.
    @params axis: Dimension along which normalization is performed.
    """
    max = arr.max(axis)
    min = arr.min(axis)
    if axis is None:
        arr -= min
        arr /= (max-min)
    else:
        arr -= np.expand_dims(min, axis=axis)
        arr /= np.expand_dims((max-min), axis=axis)

def log_normalize(arr, axis=None):
    """ Inplace Method
    Normalizes the input array so that ``sum(exp(a)) == 1``.
    @params arr : Non-normalized Input array.
    @params axis: Dimension along which normalization is performed.
    """
    with np.errstate(under="ignore"):
        arr_lse = logsumexp(arr, axis, keepdims=True)
    arr -= arr_lse

def log_mask_zero(arr):
    """Computes the log of input probabilities masking divide by zero in log."""
    arr = np.asarray(arr)
    with np.errstate(divide="ignore"):
        return np.log(arr)

# ref: https://github.com/hmmlearn/hmmlearn/blob/0e9274cb138427919c13ef79f11a7358c4e2b4a9/lib/hmmlearn/stats.py#L5
def log_multivariate_normal_density(X, means, covars, covariance_type='diag', *args, **kwargs):
    """ Compute the log probability under a multivariate Gaussian distribution.

    (PRML 2.43)
    N_i(x_n|mu_i,Sigma_i) = 1/(2*pi)^{D/2} * 1/|Sigma_i|^{1/2} * exp(-1/2*(x_n-mu_i)^T*Sigma_i^{-1}*(x_n-mu_i))

    | covariance_type | covars.shape                         |   **kwargs      |
    ----------------------------------------------------------------------------
    | 'spherical'     | (n_gaussian, n_features)             |                 |
    | 'tied'          | (n_features, n_features)             |                 |
    | 'diag'          | (n_gaussian, n_features)             |                 |
    | 'full'          | (n_gaussian, n_features, n_features) | min_covar=1.e-7 |

    @params X               : Input data. shape=(n_samples, n_features)
    @params menas           : mean vectors for n_gaussian Gaussian. shape=(n_gaussian, n_features)
    @params covars          : covariance for for n_gaussian Gaussians. The shape depends on `covariance_type`.
    @params covariance_type : Type of the covariance.
    @params *args           : No means.
    @params **kwargs        : No means.
    @return log_probs       : Log probabilities of each data point in X. shape=(n_samples, n_gaussian)
    """
    return {
        'spherical': _log_multivariate_normal_density_spherical,
        'tied': _log_multivariate_normal_density_tied,
        'diag': _log_multivariate_normal_density_diag,
        'full': _log_multivariate_normal_density_full,
    }[covariance_type](X, means, covars, *args, **kwargs)

def _log_multivariate_normal_density_spherical(X, means, covars, *args, **kwargs):
    """
    @params X        : shape=(n_samples,  n_features)
    @params menas    : shape=(n_gaussian, n_features)
    @params covars   : shape=(n_gaussian, n_features)
    @params *args    : No means.
    @params **kwargs : No means.
    @return og_probs : shape=(n_samples,  n_gaussian)
    """
    # If the input
    _,  n_features = X.shape
    cv = covars.copy()
    if covars.ndim == 1:
        # This indicates that covars.shape=(n_gaussian)
        cv = cv[:, np.newaxis]
    if cv.shape[1] == 1:
        # Assign single variance value to all diagonal elements.
        # shape=(n_gaussian, 1) → shape=(n_gaussian*1, 1*n_features)
        cv = np.tile(cv, (1, n_features))
    return _log_multivariate_normal_density_diag(X, means, cv, *args, **kwargs)

def _log_multivariate_normal_density_tied(X, means, covars, *args, **kwargs):
    """
    @params X         : shape=(n_samples,  n_features)
    @params menas     : shape=(n_gaussian, n_features)
    @params covars    : shape=(n_features, n_features)
    @params *args     : No means.
    @params **kwargs  : No means.
    @return log_probs : shape=(n_samples,  n_gaussian)
    """
    n_gaussian, _ = means.shape
    # shape=(n_features, n_features) → shape=(n_gaussian, n_features*1, n_features*1)
    cv = np.tile(covars, (n_gaussian, 1, 1))
    return _log_multivariate_normal_density_full(X, means, cv, *args, **kwargs)

def _log_multivariate_normal_density_diag(X, means, covars, *args, **kwargs):
    """
    @params X         : shape=(n_samples,  n_features)
    @params menas     : shape=(n_gaussian, n_features)
    @params covars    : shape=(n_gaussian, n_features)
    @return log_probs : shape=(n_samples,  n_gaussian)
    """
    n_samples, n_features = X.shape
    # Avoid log(0) by setting `np.finfo(float).tiny` to the smallest value.
    covars = np.maximum(covars, np.finfo(float).tiny)
    with np.errstate(over="ignore"):
        return -n_features/2*np.log(2*np.pi) \
               -1/2*np.log(covars).sum(axis=-1) + \
               -1/2*( (X[:, np.newaxis, :]-means[np.newaxis, :, :])**2 / covars[np.newaxis, :, :] ).sum(axis=-1)

def _log_multivariate_normal_density_full(X, means, covars, min_covar=1.e-7):
    """
    @params X         : shape=(n_samples,  n_features)
    @params menas     : shape=(n_gaussian, n_features)
    @params covars    : shape=(n_gaussian, n_features, n_features)
    @return log_probs : shape=(n_samples,  n_gaussian)
    """
    n_samples, n_features = X.shape
    n_gaussian, _ = means.shape
    log_prob = np.empty(shape=(n_samples, n_gaussian))
    for n, (mu, cv) in enumerate(zip(means, covars)):
        try:
            cv_chol = linalg.cholesky(cv, lower=True)
        except linalg.LinAlgError:
            # The model is most probably stuck in a component with too
            # few observations, we need to reinitialize this components
            try:
                cv_chol = linalg.cholesky(cv + min_covar*np.eye(n_features), lower=True)
            except linalg.LinAlgError:
                raise ValueError("'covars' must be symmetric, positive-definite")
        # cv_chol is lower-triangular Cholesky factor of cv.
        cv_log_det = 2*np.sum(np.log(np.diagonal(cv_chol)))
        cv_sol = linalg.solve_triangular(cv_chol, (X-mu).T, lower=True).T
        log_prob[:,n] = -1/2*(np.sum(cv_sol**2, axis=1) + n_features*np.log(2*np.pi) + cv_log_det)

    return log_prob

# ref: https://github.com/hmmlearn/hmmlearn/blob/0e9274cb138427919c13ef79f11a7358c4e2b4a9/lib/hmmlearn/utils.py#L89
def decompress_based_on_covariance_type(covars, covariance_type='full', n_gaussian=1, n_features=1):
    """
    Create the correct shape of covariance matris
    from each feature based on the covariance type.
    ~~~~~~~~
    @params covars          : compressed covariance values.
    @params covariance_type : string.
    @params n_gauusian      : number of gaussian
    @params n_features      : number of features.
    @return covariances     : covariances. shape=(n_gaussian, n_features, n_features)
    ==========================================================
    | covariance_type |               Gaussian               |
    ----------------------------------------------------------
    | "spherical"     | (n_gaussian)                         |
    | "diag"          | (n_gaussian, n_features)             |
    | "full"          | (n_gaussian, n_features, n_features) |
    | "tied"          | (n_features, n_features)             |
    """
    if covariance_type == 'full':
        return covars
    elif covariance_type == 'diag':
        # np.diag(n_features).shape=(n_features, n_features)
        # shape=(n_gaussian, n_features) → shape=(n_gaussian, (n_features, n_features))
        return np.array(list(map(np.diag, covars)))
    elif covariance_type == 'tied':
        # shape=(n_features, n_features) → shape=(n_gaussian, n_features*1, n_features*1)
        return np.tile(covars, (n_gaussian, 1, 1))
    elif covariance_type == 'spherical':
        # shape=(1, n_features, n_features)
        eye = np.eye(n_features)[np.newaxis, :, :]
        # shape=(n_gaussian, 1, 1)
        covars = covars[:, np.newaxis, np.newaxis]
        # shape=(n_gaussian, n_features, n_features)
        return eye * covars

# ref: https://github.com/hmmlearn/hmmlearn/blob/0e9274cb138427919c13ef79f11a7358c4e2b4a9/lib/hmmlearn/_utils.py#L47
def compress_based_on_covariance_type_from_tied_shape(tied_cv, covariance_type, n_gaussian):
    """ Create all the covariance matrices from a given template.
    @params tied_cv         : Input tied covariance.
    @params covariance_type : string.
    @params n_gauusian      : number of gaussian
    @return cv              : compressed values.
    ==========================================================
    | covariance_type |               Gaussian               |
    ----------------------------------------------------------
    | "spherical"     | (n_gaussian)                         |
    | "diag"          | (n_gaussian, n_features)             |
    | "full"          | (n_gaussian, n_features, n_features) |
    | "tied"          | (n_features, n_features)             |
    """
    n_features = tied_cv.shape[1]
    if covariance_type == 'spherical':
        # shape=(n_features,) → shape=(n_gaussian, n_features*1)
        # Before:
        # cv = np.tile(tied_cv.mean()*np.ones(n_features), (n_gaussian, 1))
        # Modified by Shuto Iwasaki.
        cv = np.full(shape=n_gaussian, fill_value=tied_cv.mean())
    elif covariance_type == 'tied':
        # shape=((n_features, n_features)
        cv = tied_cv
    elif covariance_type == 'diag':
        # shape=(n_features,) → shape=(n_gaussian, n_features*1)
        cv = np.tile(np.diag(tied_cv), (n_gaussian, 1))
    elif covariance_type == 'full':
        # shape=(n_features, n_features) → shape=(n_gaussian, n_features*1, n_features*1)
        cv = np.tile(tied_cv, (n_gaussian, 1, 1))
    else:
        handleKeyError(['spherical', 'tied', 'diag', 'full'], covariance_type=covariance_type)
    return cv



""" Utils for Preprocessing. """
import itertools
import numpy as np

class BaseTransformer():
    def __init__(self, _base_func_kwargs=[], **kwargs):
        """ Transformer's Base Class.
        @params _base_func_kwargs: (list) Keywards for `_base_function`.
        @params use_bias         : (bool) Whether x_transformed includes bias (all 1 array) or not.
        @params kwargs           : The pairs of 'Key' and 'Val', which are used in `_base_function`.
        """
        self.use_bias = kwargs.pop("use_bias", True)
        self._base_func_kwargs = _base_func_kwargs
        self.__dict__.update(**kwargs)
        # Preprocessing. Convert to List type to pass the `itertools`.
        for key in _base_func_kwargs:
            vals = self.__dict__[key]
            if isinstance(vals, np.ndarray):
                self.__dict__[key] = list(vals)
            elif not isinstance(vals, list):
                self.__dict__[key] = [vals]

    def _base_function(self, x, **kwargs):
        raise NotImplementedError()

    def transform(self, x):
        """
        @param  x           : (ndarray) shape=(N,D)
        @return x_tranformed: (ndarray) shape=(N,M)
        """
        if x.ndim == 1:
            x = np.expand_dims(x, axis=1)
        N,D = x.shape

        base_func_values = [self.__dict__[k] for k in self._base_func_kwargs]
        base_func_values_combs = list(itertools.product(*base_func_values))

        x_transformed = np.asarray([self._base_function(x, **dict(zip(self._base_func_kwargs,valsets))).reshape(N,-1) for valsets in base_func_values_combs])
        x_transformed = np.hstack(x_transformed)

        if self.use_bias:
            x_transformed = np.c_[np.ones(shape=(N,1)), x_transformed]

        return x_transformed

class GaussianBaseTransformer(BaseTransformer):
    def __init__(self, **kwargs):
        """ x.shape = (N,D)
        @params kwargs:
            - mu      : (list) shape=(M_mu, D)
            - sigma   : (list) shape=(M_sigma, D)
            - use_bias: (bool) Whether x_transformed includes bias or not.
                               If True, x_transformed.shape[1] += 1.
        @return x_tranformed: shape=(N, M_mu*M_sigma)
        """
        super().__init__(_base_func_kwargs=["mu", "sigma"], **kwargs)

    def _base_function(self, x, mu, sigma):
        """
        @param x:  (ndarray) shape=(N,D)
        @param mu: (ndarray) shape=(D,)
        """
        return np.exp(-1/2 * np.sum(np.square(x-mu), axis=-1) / sigma**2)

class SigmoidBaseTransformer(BaseTransformer):
    def __init__(self, **kwargs):
        """ x.shape = (N,D)
        @params kwargs:
            - mu      : (list) shape=(M_mu, D)
            - sigma   : (list) shape=(M_sigma, D)
            - use_bias: (bool) Whether x_transformed includes bias or not.
                               If True, x_transformed.shape[1] += 1.
        @return x_tranformed: shape=(N, M_mu*M_sigma)
        """
        super().__init__(_base_func_kwargs=["mu", "sigma"], **kwargs)

    def _base_function(self, x, mu, sigma):
        """
        @param x:  (ndarray) shape=(N,D)
        @param mu: (ndarray) shape=(D,)
        """
        return 1 / (1 + np.exp(- (x-mu)/sigma ))

class PolynomialBaseTransformer(BaseTransformer):
    def __init__(self, **kwargs):
        """ x.shape = (N,D)
        @params kwargs:
            - exponent: (list) shape=(M_exponent) each element represents the exponent.
            - use_bias: (bool) Whether x_transformed includes bias or not.
                               If True, x_transformed.shape[1] += 1.
        @return x_tranformed: shape=(N, M)
        """
        super().__init__(_base_func_kwargs=["exponent"], **kwargs)

    def _base_function(self, x, exponent):
        """
        @param exponent: int
        """
        return x**exponent

class NoneBaseTransformer(BaseTransformer):
    def __init__(self, **kwargs):
        super().__init__(_base_func_kwargs=[], **kwargs)

    def _base_function(self, x):
        return x

BasisTransformerHandler = {
    'gaussian': GaussianBaseTransformer,
    'sigmoid': SigmoidBaseTransformer,
    'polynomial': PolynomialBaseTransformer,
    'none': NoneBaseTransformer,
}

def basis_transformer(basis_name, **basisargs):
    if basis_name not in BasisTransformerHandler:
        raise KeyError(f"Please select basis transformer from {list(BasisTransformerHandler.keys())}")
    return BasisTransformerHandler[basis_name](**basisargs)



# coding: utf-8
from __future__ import absolute_import

import numpy as np

from ..activations import ActivationFunc
from ..initializers import Initializer
from ..losses import LossFunc
from ..engine.base_layer import Layer
from ..utils import flush_progress_bar

class Input(Layer):
    def __init__(self, input_shape, **kwargs):
        self.input_shape=input_shape
        super().__init__(**kwargs)

    def forward(self, input):
        return input

    def backprop(self, delta):
        return delta

class Flatten(Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def compute_output_shape(self, input_shape):
        self.input_shape = input_shape
        self.output_shape = (np.prod(list(input_shape)),)
        return self.output_shape

    def forward(self, input):
        return input.flatten()

    def backprop(self, delta):
        return delta.reshape(self.input_shape)

class Dense(Layer):
    def __init__(self, units, activation='linear', kernel_initializer='random_normal', bias_initializer='zeros', **kwargs):
        """
        @param units             : (tuple) dimensionality of the (input space, output space).
        @param activation        : (str) Activation function to use.
        @param kernel_initializer: (str) Initializer for the `kernel` weights matrix.
        @param bias_initializer  : (str) Initializer for the bias vector.
        """
        self.output_shape=(units,)
        self.kernel_initializer = Initializer(kernel_initializer)
        self.kernel_regularizer = None
        self.kernel_constraint  = None
        self.bias_initializer   = Initializer(bias_initializer)
        self.bias_regularizer   = None
        self.bias_constraint    = None
        self.h = ActivationFunc(activation)
        self.use_bias = True
        super().__init__(**kwargs)

    def compute_output_shape(self, input_shape):
        return self.output_shape

    def build(self, input_shape):
        self.input_shape = input_shape
        output_shape = self.compute_output_shape(input_shape)
        self.kernel  = self.add_weight(shape=(self.output_shape + input_shape),
                                       name="kernel",
                                       initializer=self.kernel_initializer,
                                       regularizer=self.kernel_regularizer,
                                       constraint =self.kernel_constraint)
        if self.use_bias:
            self.bias = self.add_weight(shape=(self.output_shape[0],1),
                                        name="bias",
                                        initializer=self.bias_initializer,
                                        regularizer=self.bias_regularizer,
                                        constraint =self.bias_constraint)
        else:
            self.bias = np.empty(shape=(self.output_shape[0],1))
        return output_shape

    def forward(self, input):
        """ @param input: shape=(Din,) """
        Xin  = np.append(input,1) if self.use_bias else input # shape=(Din+1,)
        a    = np.c_[self.kernel, self.bias].dot(Xin) # (Dout,Din+1) @ (Din+1,) = (Dout,)
        Xout = self.h.forward(input=a) # shape=(Dout,)
        self.a = a
        self.Xin = Xin # shape=(Din+1,) or shape=(Din,)
        return Xout

    def backprop(self, dEdXout):
        """ @param dEdXout: shape=(Dout,) """
        # dXoutda = self.h.diff(self.a) # shape=(Dout,)
        # dEda = dEdXout.dot(dXoutda) if dXoutda.ndim==2 else dEdXout * dXoutda # shape=(Dout,)
        dEda = self.h.diff(self.a) * dEdXout
        dEdXin = np.c_[self.kernel, self.bias].T.dot(dEda) # (Din+1,Dout) @ (Dout,) = (Din+1,)
        if self.trainable:
            self.memorize_delta(dEda)
        dEdXin = dEdXin[:-1] if self.use_bias else dEdXin
        return dEdXin # shape=(Din,) delta of bias is not propagated.

    def memorize_delta(self, dEda):
        dEdw = np.outer(dEda, self.Xin)
        if self.use_bias:
            self._losses['kernel'] += dEdw[:,:-1] # shape=(Dout, Din)
            self._losses['bias'] += dEdw[:,-1:] # shape=(Dout, 1)
        else:
            self._losses['kernel'] += dEdw

class Dropout(Layer):
    def __init__(self, keep_prob, **kwargs):
        self.keep_prob = min(1., max(0., keep_prob))
        super().__init__(**kwargs)

    def compute_output_shape(self, input_shape):
        self.input_shape = input_shape
        self.output_shape = input_shape
        return input_shape

    def forward(self, input):
        self.mask = np.random.uniform(low=0., high=1., size=input.shape)<=self.keep_prob
        return input * self.mask

    def backprop(self, delta):
        return delta * self.mask



# coding: utf-8
# ref: https://packaging.python.org/tutorials/packaging-projects/

import setuptools

VERSION = "0.0.0"
DESCRIPTION = \
'A set of python modules for machine learning and data mining \
especially in the biological field.'

def setup_package():
    metadata = dict(
        name='kerasy',
        version=VERSION,
        description=DESCRIPTION,
        author='Shuto Iwasaki',
        author_email='cabernet.rock@gmail.com',
        license='MIT',
        url='https://iwasakishuto.github.io/Kerasy/doc',
        project_urls={
            'Documentation': 'https://iwasakishuto.github.io/Kerasy/doc/index.html',
            'Source Code': 'https://github.com/iwasakishuto/Kerasy',
        },
        packages=setuptools.find_packages(),
        python_requires=">=3.6",
        install_requires=[
            'numpy>=1.15.1',
            'scipy>=1.4.1',
            'seaborn>=0.10.0',
            'Cython>=0.28.5',
        ],
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
            'Topic :: Software Development :: Libraries',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    )
    setuptools.setup(**metadata)

if __name__ == "__main__":
    setup_package()



from __future__ import absolute_import
import os

UTILS_DIR_PATH = os.path.dirname(os.path.abspath(__file__))

from . import bio_utils
from . import deep_utils
from . import generic_utils
from . import metric_utils
from . import model_select_utils
from . import monitor_utils
from . import np_utils
from . import param_utils
from . import prepro_utils
from . import toydata_utils
from . import training_utils
from . import tree_utils
from . import vis_utils

# Globally-importable utils.
from .bio_utils import bpHandler
from .bio_utils import arangeFor_printAlignment
from .bio_utils import printAlignment
from .bio_utils import readMonoSeq
from .bio_utils import readMultiSeq

from .deep_utils import get_params_size
from .deep_utils import print_summary

from .generic_utils import flatten_dual
from .generic_utils import get_varname
from .generic_utils import disp_var_globals
from .generic_utils import disp_val_globals
from .generic_utils import get_file
from .generic_utils import priColor
from .generic_utils import handleKeyError
from .generic_utils import handleTypeError
from .generic_utils import urlDecorate
from .generic_utils import measure_complexity
from .generic_utils import has_all_attrs
from .generic_utils import has_not_attrs
from .generic_utils import handle_random_state
from .generic_utils import fout_args
from .generic_utils import format_spec_create
from .generic_utils import print_func_create
from .generic_utils import Table

from .metric_utils import norm_vectors
from .metric_utils import normalize_vectors
from .metric_utils import mean_squared_error
from .metric_utils import root_mean_squared_error
from .metric_utils import silhouette_samples
from .metric_utils import silhouette_score
from .metric_utils import check_paired_array
from .metric_utils import paired_distances
from .metric_utils import paired_euclidean_distances
from .metric_utils import paired_manhattan_distances
from .metric_utils import paired_cosine_distances
from .metric_utils import check_pairwise_array
from .metric_utils import pairwise_distances
from .metric_utils import pairwise_euclidean_distances

from .model_select_utils import cross_validation

from .monitor_utils import ConvergenceMonitor
from .monitor_utils import ThresholdMonitor

from .np_utils import CategoricalEncoder
from .np_utils import findLowerUpper
from .np_utils import inverse_arr
from .np_utils import _check_sample_weight
from .np_utils import normalize
from .np_utils import standardize
from .np_utils import log_normalize
from .np_utils import log_mask_zero
from .np_utils import log_multivariate_normal_density
from .np_utils import compress_based_on_covariance_type_from_tied_shape
from .np_utils import decompress_based_on_covariance_type

from .param_utils import Params

from .prepro_utils import basis_transformer

from .progress_utils import flush_progress_bar
from .progress_utils import flush_progress_plot

from .toydata_utils import generateX
from .toydata_utils import generateSin
from .toydata_utils import generateGausian
from .toydata_utils import generateMultivariateNormal
from .toydata_utils import generateWholeCakes
from .toydata_utils import generateWhirlpool
from .toydata_utils import generateSeq
from .toydata_utils import generateSeq_embedded_Motif

from .training_utils import make_batches
from .training_utils import train_test_split
from .training_utils import iter_from_variable_len_samples

from .tree_utils import DecisionTreeDOTexporter
from .tree_utils import ItemsetTreeDOTexporter

from .vis_utils import rgb2hex, rgba2rgb, hex2rgb
from .vis_utils import chooseTextColor
from .vis_utils import mk_color_dict
from .vis_utils import measureCanvas
from .vis_utils import galleryplot
from .vis_utils import plot_2D
from .vis_utils import objVSexp
from .vis_utils import silhouette_plot
from .vis_utils import _left_silhouette_plot
from .vis_utils import _right_silhouette_plot
from .vis_utils import plot_CDF
from .vis_utils import plot_roc_curve



# coding: utf-8
import numpy as np

def get_params_size(layer):
    trainable_params = np.sum([layer.__dict__.get(key).size for key in layer._trainable_weights]).astype(int)
    not_trainable_params = np.sum([layer.__dict__.get(key).size for key in layer._non_trainable_weights]).astype(int)
    total_params = trainable_params+not_trainable_params
    return (total_params, trainable_params, not_trainable_params)

def print_summary(self):
    col_widths = (28,25,9)
    w1,w2,w3 = col_widths
    line_length = sum(col_widths)+len(col_widths)

    def arange_row_content(layer_type, output_shape, n_params):
        return f"{str(layer_type)[:w1]:<{w1}} {str(output_shape)[:w2]:<{w2}} {str(n_params)[:w3]:<{w3}} "

    print("-"*line_length)
    print(arange_row_content("Layer (type)", "Output Shape", "Param #"))
    print("="*line_length)

    tots = ables = nons = 0
    for i,layer in enumerate(self.layers):
        if i>0: print("-"*line_length)
        tot, able, non = get_params_size(layer)
        tots+=tot; ables+=able; nons+=non
        print(arange_row_content(
            f"{layer.name} ({layer.__class__.__name__})",
            str((None,) + layer.output_shape),
            tot,
        ))

    print("="*line_length)
    print(f"Total params: {tots:,}")
    print(f"Trainable params: {ables:,}")
    print(f"Non-trainable params: {nons:,}")
    print("-"*line_length)



# coding: utf-8
from .generic_utils import handleKeyError
from .vis_utils import mk_color_dict
from .vis_utils import chooseTextColor
from .vis_utils import rgb2hex

BLACK_hex = "#000000"
WHITE_hex = "#FFFFFF"

TREE_STRUCTURES = ["lr", "sons"]

class BaseTreeDOTexporter():
    def __init__(self, feature_names=None, class_names=None, cmap="jet",
                 filled=True, rounded=True, precision=3, tree_structure="lr"):
        self.node_id = None
        self.dot_text = None
        self.feature_names = feature_names
        self.class_names = class_names
        self.filled = filled
        self.rounded = rounded
        self.precision = precision
        handleKeyError(lst=TREE_STRUCTURES, tree_structure=tree_structure)
        self.tree_structure = tree_structure
        # Coloring according to `class names`.
        if class_names is not None:
            self._mk_color_dict(class_names=class_names, cmap=cmap)

    def _mk_color_dict(self, class_names, cmap):
        bg_color_dict = mk_color_dict(class_names, cmap=cmap, ctype="rgb")
        fw_color_dict = {
            cls: rgb2hex(chooseTextColor(rgb_bg)) for cls,rgb_bg in bg_color_dict.items()
        }
        bg_color_dict = {cls : rgb2hex(rgb_bg) for cls,rgb_bg in bg_color_dict.items()}
        if not self.filled:
            fw_color_dict = {cls : BLACK_hex for cls,rgb_bg in fw_color_dict.items()}
        self.fill_color_dict = bg_color_dict
        self.font_color_dict = fw_color_dict

    def _add_head(self):
        head_info = "digraph Tree {\nnode [shape=box"
        # Specify node aesthetics (`round`, and `filled`).
        rounded_filled = []
        rounded_filled.extend(["filled"]  if self.filled  else [])
        rounded_filled.extend(["rounded"] if self.rounded else [])
        if len(rounded_filled) > 0:
            head_info += f", style=\"{', '.join(rounded_filled)}\", color=\"black\""
        if self.rounded:
            head_info += ", fontname=helvetica"
        head_info += "] ;\n"
        # Specify graph & edge aesthetics
        if self.rounded:
            head_info += "\tedge [fontname=helvetica] ;\n"
        self.dot_text += head_info

    def _add_node_info(self, node_id, label_="", fillcolor_=WHITE_hex, fontcolor_=BLACK_hex):
        self.dot_text += f'\t{node_id} [label=<{label_}>, fontcolor="{fontcolor_}", fillcolor="{fillcolor_}"] ;\n'

    def _add_par_chil_info(self, son_id, par_id, arrow_info_=""):
        self.dot_text += f'\t{par_id} -> {son_id} {arrow_info_} ;\n'

    def _add_tail(self):
        self.dot_text += "}"

    def _initialize(self):
        self.node_id = 0
        self.dot_text = ""

    def recurse(self, node, par_node_id=None):
        {
            "lr"   : self._recurse_lr,
            "sons" : self._recurse_sons,
        }[self.tree_structure](node=node, par_node_id=par_node_id)

    def _recurse_lr(self, node, par_node_id=None):
        my_node_id = self.node_id
        self.node_id += 1
        self._add_node_info(node=node, node_id=my_node_id)
        if par_node_id is not None:
            self._add_par_chil_info(son_id=my_node_id, par_id=par_node_id)
        if node.left is not None:
            self._recurse_lr(node.left, my_node_id)
        if node.right is not None:
            self._recurse_lr(node.right, my_node_id)

    def _recurse_sons(self, node, par_node_id=None):
        my_node_id = self.node_id
        self.node_id += 1
        self._add_node_info(node=node, node_id=my_node_id)
        if par_node_id is not None:
            self._add_par_chil_info(son_id=my_node_id, par_id=par_node_id)
        for son in node.sons:
            self._recurse_sons(son, my_node_id)

    def export(self, node, out_file=None):
        self._initialize()
        self._add_head()
        self.recurse(node, par_node_id=None)
        self._add_tail()
        if out_file is not None:
            with open(out_file, mode="w", encoding="utf-8") as f:
                f.write(self.dot_text)
            print(f"{out_file} was created.")
        else:
            return self.dot_text

    def write_png(self, node, path):
        try:
            import pydotplus
        except:
            raise ImportError("You Need to run the following command\
                              \n`$ pip install -U pydotplus`")
        if path[-4:] != ".png":
            path += ".png"
        dot_data = self.export(node, out_file=None)
        graph = pydotplus.graph_from_dot_data(dot_data)
        return graph.write_png(path, f='png', prog='dot')

class DecisionTreeDOTexporter(BaseTreeDOTexporter):
    def __init__(self, feature_names, class_names, cmap="jet",
                 filled=True, rounded=True, precision=3):
        super().__init__(feature_names=feature_names, class_names=class_names,
                         cmap=cmap, filled=filled, rounded=rounded,
                         precision=precision, tree_structure="lr")
        self.headlabeled = False

    def _initialize(self):
        self.headlabeled = False
        super()._initialize()

    def _add_node_info(self, node, node_id):
        precision = self.precision
        class_names = self.class_names
        # label info.
        label = ""
        if not (node.feature == None or node.depth == node.max_depth):
            label += f"{self.feature_names[node.feature]} &le; {round(node.threshold, precision)}<br/>"
        label += f"{node.criterion} = {round(node.impurity, precision)}<br/>" \
               + f"N = {node.num_samples}<br/>" \
               + f"classes = {node.num_classes}<br/>" \
               + f"class = {class_names[node.label]}"
        # fill color info.
        fillcolor = self.fill_color_dict.get(class_names[node.label], WHITE_hex)
        # font color info.
        fontcolor = self.font_color_dict.get(class_names[node.label], BLACK_hex)
        super()._add_node_info(node_id=node_id, label_=label, fillcolor_=fillcolor, fontcolor_=fontcolor)

    def _add_par_chil_info(self, son_id, par_id):
        arrow_info_ = ""
        if par_id==0:
            deg = "" if self.headlabeled else "-"
            arrow_info_ += f'[labeldistance=2.5, labelangle={deg}45, headlabel="{not self.headlabeled}"]'
            self.headlabeled = not self.headlabeled
        super()._add_par_chil_info(son_id=son_id, par_id=par_id, arrow_info_=arrow_info_)

class ItemsetTreeDOTexporter(BaseTreeDOTexporter):
    def __init__(self, class_names, cmap="jet",
                 filled=True, rounded=True, precision=3):
        super().__init__(feature_names=None, class_names=class_names,
                         cmap=cmap, filled=filled, rounded=rounded,
                         precision=precision, tree_structure="sons")

    def _add_node_info(self, node, node_id):
        # label info.
        label = ""
        label += f"frequency = {node.freq}<br/>" \
               + f"itemset = {[self.class_names[idx] for idx in node.itemset]}<br/>"
        super()._add_node_info(node_id=node_id, label_=label)

    # def _add_par_chil_info(self, son_id, par_id):
    #     arrow_info_ = ""
    #     if par_id==0:
    #         deg = "" if self.headlabeled else "-"
    #         arrow_info_ += f'[labeldistance=2.5, labelangle={deg}45, headlabel="{not self.headlabeled}"]'
    #         self.headlabeled = not self.headlabeled
    #     super()._add_par_chil_info(son_id=son_id, par_id=par_id, arrow_info_=arrow_info_)



# coding: utf-8
import numpy as np
import matplotlib.pyplot as plt
from ..utils import flush_progress_bar

class Pannel():
    n=None;m=None;digit=None
    goal=None

    def __init__(self,string,par_string,h,heuristic_method="Manhattan_distance",n=None,m=None,goal_str=None):
        """
        @param string           : (str) encoded pannel states.
        @param par_string       : (str) encoded parent pannel states. It is neccesary for traceback.
        @param h                : (int) Transition counts from initial states.
        @param heuristic_method : (str) How to estimate the total transition costs.
        @param n,m              : (int) pannel size (n × m)
        @param goal_str         : (str) encoded last pannel states.
        """
        if n is not None:
            Pannel.n = n
            Pannel.m = m
            Pannel.digit = len(str(n*m))
            Pannel.goal = goal_str

        self.str = string
        self.h = h
        self.hm = heuristic_method
        self.g = self.heuristic(heuristic_method)
        self.par = par_string

    def __lt__(self, other):
        """ this enables us to compare the pannel directoly. """
        return (self.h+self.g)<(other.h+other.g)

    def arr2str(self, array):  return ",".join(array.flatten().astype(str))
    def str2arr(self, string): return np.asarray([int(e) for e in string.split(",")]).reshape(Pannel.n,Pannel.m)

    def swap(self,array,xi,yi,xj,yj):
        """ array[xi,yi] ↔︎ array[xj,yj] """
        array = np.copy(array)
        tmp = array[xi,yi]
        array[xi,yi]=array[xj,yj]
        array[xj,yj]=tmp
        return self.arr2str(array) # NOTE: type(return) is str.

    def heuristic(self, method):
        """ Manhattan distance """
        if method=="Manhattan_distance":
            g=0
            goal_arr = self.str2arr(Pannel.goal)
            arr = self.str2arr(self.str)
            for i in range(1,Pannel.n*Pannel.m+1):
                idx = np.argmax(arr==i)
                x,y = (idx//Pannel.m, idx%Pannel.m)
                goal_idx = np.argmax(goal_arr==i)
                goal_x,goal_y = (goal_idx//Pannel.m, goal_idx%Pannel.m)
                g += abs(goal_x-x)+abs(goal_y-y)

        elif method=="Absolute_distance":
            g = sum([ei!=ej for ei,ej in zip(self.str.split(","), Pannel.goal.split(","))])
        else:
            raise KeyError(f"{method} is not defined.")
        return g

    def transibles(self):
        arr = self.str2arr(self.str)
        idx = np.argmax(arr==-1)
        x,y = (idx//Pannel.m, idx%Pannel.m)

        cands=[]
        if x!=0: cands.append(self.swap(arr,x,y,x-1,y))
        if y!=0: cands.append(self.swap(arr,x,y,x,y-1))
        if x!=Pannel.n-1: cands.append(self.swap(arr,x,y,x+1,y))
        if y!=Pannel.m-1: cands.append(self.swap(arr,x,y,x,y+1))

        return [Pannel(string=c_str, par_string=self.str, h=self.h+1, heuristic_method=self.hm) for c_str in cands]

    def plot(self, ax=None):
        if ax==None:
            fig,ax=plt.subplots()

        arr = self.str2arr(self.str)
        matrix= np.asarray(arr==-1, dtype=int)
        ax.matshow(matrix, alpha=0.3, vmin=0, vmax=1,cmap="Greys")
        for i in range(Pannel.n):
            for j in range(Pannel.m):
                text = "" if arr[i][j]<0 else arr[i][j]
                ax.text(x=j, y=i, s=text, va='center', ha='center')
        ax.set_xticks([]), ax.set_yticks([])
        return ax

def OptimalTransit(n,m,initial_str,last_str,heuristic_method="Manhattan_distance",n_row=5):
    initial_state = Pannel(
        string=initial_str,
        par_string=None,
        h=0,
        heuristic_method=heuristic_method,
        n=n,
        m=m,
        goal_str=last_str,
    )

    open_list = [initial_state]
    closed_list = {}
    it = 0; max_it = 100
    while(open_list):
        state = open_list.pop(np.argmin(open_list))
        closed_list[state.str] = state
        if state.str==last_str: break
        open_list += [s for s in state.transibles() if s.str not in closed_list]
        it+=1
        flush_progress_bar(it, max_it, metrics=f"{state.h} transition from initial states.", barname="")
        if it==max_it: it=0
        
    if last_str not in closed_list:
        print("Can't reach to the last state.")
    else:
        n_transition = closed_list[last_str].h
        print(f"n_transition = {n_transition}")

        # Trace Back.
        pannel_str = last_str
        pannel_lst = []
        while pannel_str:
            state = closed_list[pannel_str]
            pannel_str = state.par
            pannel_lst.append(state)

        n_fig = n_transition+1
        n_col = n_fig//n_row if n_fig%n_row==0 else n_fig//n_row+1

        fig = plt.figure(figsize=(4*n_row,4*n_col))
        for i, state in enumerate(pannel_lst):
            ax = fig.add_subplot(n_col,n_row, i+1)
            state.plot(ax=ax)
            ax.set_title(f"No.{i:<0{len(str(n_transition))}}")



# coding: utf-8
import numpy as np

def cross_validation(k, x, y, metrics, modelcls, shuffle=True, seed=None, modelkwargs={}, trainkwargs={}):
    """
    @params k        : k-fold
    @params x        : Explanatory variables.
    @params y        : Objective variables.
    @params metrics  : (func) How to measure prediction error.
    @params modelcls : Model Class. We will create k instances.
    @params modelargs: Model instance's argments.
    """
    n_samples = len(x)
    idx = np.repeat(np.arange(k), np.round(n_samples/k))
    if shuffle: np.random.RandomState(seed).shuffle(idx)

    score = 0
    for cls in range(k):
        x_train = x[idx==cls]; x_test = x[idx!=cls]
        y_train = y[idx==cls]; y_test = y[idx!=cls]
        model = modelcls(**modelkwargs)
        model.fit(x_train, y_train, **trainkwargs)
        y_pred = model.predict(x_test)
        score += metrics(y_test, y_pred)
    return score/k
